from __future__ import annotations

from pathlib import Path
from textwrap import dedent
from typing import TYPE_CHECKING

import pytest
from pytest_lazy_fixtures.lazy_fixture import lf as lazy_fixture

from semantic_release.hvcs.github import Github

from tests.const import (
    MAIN_PROG_NAME,
    VERSION_SUBCMD,
)
from tests.fixtures.repos import (
    repo_w_no_tags_conventional_commits,
    repo_w_trunk_only_conventional_commits,
)
from tests.util import assert_successful_exit_code

if TYPE_CHECKING:
    from unittest.mock import MagicMock

    from git import Repo
    from requests_mock import Mocker

    from tests.conftest import RunCliFn
    from tests.e2e.conftest import StripLoggingMessagesFn
    from tests.fixtures.example_project import GetWheelFileFn, UpdatePyprojectTomlFn
    from tests.fixtures.git_repo import BuiltRepoResult, GetVersionsFromRepoBuildDefFn


# No-op shouldn't change based on the branching/merging of the repository
@pytest.mark.parametrize(
    "repo_result, next_release_version",
    # must use a repo that is ready for a release to prevent no release
    # logic from being triggered before the noop logic
    [(lazy_fixture(repo_w_no_tags_conventional_commits.__name__), "1.0.0")],
)
def test_version_noop_is_noop(
    repo_result: BuiltRepoResult,
    next_release_version: str,
    run_cli: RunCliFn,
    mocked_git_fetch: MagicMock,
    mocked_git_push: MagicMock,
    post_mocker: Mocker,
    get_wheel_file: GetWheelFileFn,
):
    repo: Repo = repo_result["repo"]
    build_result_file = get_wheel_file(next_release_version)

    # Setup: reset any uncommitted changes (if any)
    repo.git.reset("--hard")

    # Setup: take measurement before running the version command
    head_sha_before = repo.head.commit.hexsha
    tags_before = {tag.name for tag in repo.tags}

    # Act
    cli_cmd = [MAIN_PROG_NAME, "--noop", VERSION_SUBCMD]
    result = run_cli(cli_cmd[1:])

    # take measurement after running the version command
    head_after = repo.head.commit
    tags_after = {tag.name for tag in repo.tags}
    tags_set_difference = set.difference(tags_after, tags_before)

    # Evaluate (no release actions should have occurred when no bump)
    assert_successful_exit_code(result, cli_cmd)
    assert f"{next_release_version}\n" == result.stdout

    # No commit has been made
    assert head_sha_before == head_after.hexsha, "HEAD should not have changed"
    assert not tags_set_difference  # No tag created

    # no build result
    assert not build_result_file.exists()

    # no file changes (since no commit was made then just check for non-committed changes)
    assert not repo.git.status(short=True)

    assert mocked_git_push.call_count == 0  # no git push of tag or commit
    assert post_mocker.call_count == 0  # no vcs release


@pytest.mark.parametrize(
    "repo_result",
    [lazy_fixture(repo_w_trunk_only_conventional_commits.__name__)],
)
def test_version_no_git_verify(
    repo_result: BuiltRepoResult,
    run_cli: RunCliFn,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    mocked_git_fetch: MagicMock,
    mocked_git_push: MagicMock,
    post_mocker: Mocker,
):
    repo = repo_result["repo"]

    # setup: set configuration setting
    update_pyproject_toml("tool.semantic_release.no_git_verify", True)
    repo.git.commit(
        m="chore: adjust project configuration for --no-verify release commits", a=True
    )
    # Fake an automated push to remote by updating the remote tracking branch
    repo.git.update_ref(
        f"refs/remotes/origin/{repo.active_branch.name}",
        repo.head.commit.hexsha,
    )

    # setup: create executable pre-commit script
    precommit_hook = Path(repo.git_dir, "hooks", "pre-commit")
    precommit_hook.parent.mkdir(parents=True, exist_ok=True)
    precommit_hook.write_text(
        dedent(
            """\
            #!/bin/sh
            echo >&2 "Always fail pre-commit" && exit 1;
            """
        )
    )
    precommit_hook.chmod(0o754)

    # setup: set git configuration to have the pre-commit hook
    repo.git.config(
        "core.hookspath",
        str(precommit_hook.parent.relative_to(repo.working_dir)),
        local=True,
    )

    # Take measurement beforehand
    head_sha_before = repo.head.commit.hexsha
    tags_before = {tag.name for tag in repo.tags}

    # Execute
    cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, "--patch"]
    result = run_cli(cli_cmd[1:])

    # Take measurement after the command
    head_after = repo.head.commit
    tags_after = {tag.name for tag in repo.tags}
    tags_set_difference = set.difference(tags_after, tags_before)

    # Evaluate (normal release actions should have occurred when forced patch bump)
    assert_successful_exit_code(result, cli_cmd)
    # A commit has been made (regardless of precommit)
    assert [head_sha_before] == [head.hexsha for head in head_after.parents]
    assert len(tags_set_difference) == 1  # A tag has been created
    assert mocked_git_fetch.call_count == 1  # fetch called to check for remote changes
    assert mocked_git_push.call_count == 2  # 1 for commit, 1 for tag
    assert post_mocker.call_count == 1  # vcs release creation occurred


@pytest.mark.parametrize(
    "repo_result", [lazy_fixture(repo_w_trunk_only_conventional_commits.__name__)]
)
def test_version_on_nonrelease_branch(
    repo_result: BuiltRepoResult,
    run_cli: RunCliFn,
    mocked_git_fetch: MagicMock,
    mocked_git_push: MagicMock,
    post_mocker: Mocker,
    strip_logging_messages: StripLoggingMessagesFn,
):
    """
    Given repo is on a non-release branch,
    When running the version command,
    Then no version release should happen which means no code changes, no build, no commit,
    no tag, no push, and no vcs release creation while returning a successful exit code
    """
    repo = repo_result["repo"]

    branch = repo.create_head("next").checkout()
    expected_error_msg = (
        f"branch '{branch.name}' isn't in any release groups; no release will be made\n"
    )
    repo_status_before = repo.git.status(short=True)
    head_before = repo.head.commit.hexsha
    tags_before = sorted([tag.name for tag in repo.tags])

    # Act
    cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD]
    result = run_cli(cli_cmd[1:])

    # Evaluate (expected -> actual)
    assert_successful_exit_code(result, cli_cmd)
    assert not result.stdout
    assert expected_error_msg == strip_logging_messages(result.stderr)

    # assert nothing else happened (no code changes, no commit, no tag, no push, no vcs release)
    tags_after = sorted([tag.name for tag in repo.tags])
    assert repo_status_before == repo.git.status(short=True)
    assert head_before == repo.head.commit.hexsha
    assert tags_before == tags_after
    assert mocked_git_push.call_count == 0
    assert post_mocker.call_count == 0


@pytest.mark.parametrize(
    "repo_result",
    [lazy_fixture(repo_w_trunk_only_conventional_commits.__name__)],
)
def test_version_on_last_release(
    repo_result: BuiltRepoResult,
    get_versions_from_repo_build_def: GetVersionsFromRepoBuildDefFn,
    run_cli: RunCliFn,
    mocked_git_fetch: MagicMock,
    mocked_git_push: MagicMock,
    post_mocker: Mocker,
    strip_logging_messages: StripLoggingMessagesFn,
):
    """
    Given repo is on the last release version,
    When running the version command,
    Then no version release should happen which means no code changes, no build, no commit,
    no tag, no push, and no vcs release creation while returning a successful exit code and
    printing the last release version
    """
    repo = repo_result["repo"]
    latest_release_version = get_versions_from_repo_build_def(
        repo_result["definition"]
    )[-1]
    expected_error_msg = (
        f"No release will be made, {latest_release_version} has already been released!"
    )

    # Setup: take measurement before running the version command
    repo_status_before = repo.git.status(short=True)
    head_before = repo.head.commit.hexsha
    tags_before = sorted([tag.name for tag in repo.tags])

    # Act
    cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD]
    result = run_cli(cli_cmd[1:], env={Github.DEFAULT_ENV_TOKEN_NAME: "1234"})

    # take measurement after running the version command
    repo_status_after = repo.git.status(short=True)
    head_after = repo.head.commit.hexsha
    tags_after = sorted([tag.name for tag in repo.tags])

    # Evaluate (expected -> actual)
    assert_successful_exit_code(result, cli_cmd)
    assert f"{latest_release_version}\n" == result.stdout
    assert f"{expected_error_msg}\n" == strip_logging_messages(result.stderr)

    # assert nothing else happened (no code changes, no commit, no tag, no push, no vcs release)
    assert repo_status_before == repo_status_after
    assert head_before == head_after
    assert tags_before == tags_after
    assert mocked_git_push.call_count == 0
    assert post_mocker.call_count == 0


@pytest.mark.parametrize(
    "repo_result", [lazy_fixture(repo_w_no_tags_conventional_commits.__name__)]
)
def test_version_only_tag_push(
    repo_result: BuiltRepoResult,
    run_cli: RunCliFn,
    mocked_git_fetch: MagicMock,
    mocked_git_push: MagicMock,
    post_mocker: Mocker,
) -> None:
    """
    Given a repo with no tags,
    When running the version command with the `--no-commit` and `--tag` flags,
    Then a tag should be created on the current commit, pushed, and a release created.
    """
    repo = repo_result["repo"]

    # Setup
    head_before = repo.head.commit

    # Act
    cli_cmd = [
        MAIN_PROG_NAME,
        VERSION_SUBCMD,
        "--no-commit",
        "--tag",
    ]
    result = run_cli(cli_cmd[1:])

    # capture values after the command
    tag_after = repo.tags[-1].name
    head_after = repo.head.commit

    # Assert only tag was created, it was pushed and then release was created
    assert_successful_exit_code(result, cli_cmd)
    assert tag_after == "v1.0.0"
    assert head_before == head_after
    assert mocked_git_push.call_count == 1  # 0 for commit, 1 for tag
    assert post_mocker.call_count == 1



from semantic_release.commit_parser.conventional.options import (
    ConventionalCommitParserOptions,
)
from semantic_release.commit_parser.conventional.options_monorepo import (
    ConventionalCommitMonorepoParserOptions,
)
from semantic_release.commit_parser.conventional.parser import ConventionalCommitParser
from semantic_release.commit_parser.conventional.parser_monorepo import (
    ConventionalCommitMonorepoParser,
)

__all__ = [
    "ConventionalCommitParser",
    "ConventionalCommitParserOptions",
    "ConventionalCommitMonorepoParser",
    "ConventionalCommitMonorepoParserOptions",
]



from __future__ import annotations

import logging
import os
from collections.abc import Mapping
from dataclasses import dataclass, is_dataclass
from enum import Enum
from functools import reduce
from pathlib import Path
from re import (
    Pattern,
    compile as regexp,
    error as RegExpError,  # noqa: N812
    escape as regex_escape,
)
from typing import Any, ClassVar, Dict, List, Literal, Optional, Tuple, Type, Union

# typing_extensions is for Python 3.8, 3.9, 3.10 compatibility
import tomlkit
from git import Actor, InvalidGitRepositoryError
from git.repo.base import Repo
from jinja2 import Environment
from pydantic import (
    BaseModel,
    Field,
    RootModel,
    ValidationError,
    field_validator,
    model_validator,
)
from typing_extensions import Annotated, Self
from urllib3.util.url import parse_url

import semantic_release.hvcs as hvcs
from semantic_release.changelog.context import ChangelogMode
from semantic_release.changelog.template import environment
from semantic_release.cli.const import DEFAULT_CONFIG_FILE
from semantic_release.cli.masking_filter import MaskingFilter
from semantic_release.commit_parser import (
    AngularCommitParser,
    CommitParser,
    ConventionalCommitMonorepoParser,
    ConventionalCommitParser,
    EmojiCommitParser,
    ParseResult,
    ParserOptions,
    ScipyCommitParser,
    TagCommitParser,
)
from semantic_release.const import COMMIT_MESSAGE, DEFAULT_COMMIT_AUTHOR
from semantic_release.errors import (
    DetachedHeadGitError,
    InvalidConfiguration,
    MissingGitRemote,
    NotAReleaseBranch,
    ParserLoadError,
)
from semantic_release.globals import logger
from semantic_release.helpers import dynamic_import
from semantic_release.version.declarations.i_version_replacer import IVersionReplacer
from semantic_release.version.declarations.pattern import PatternVersionDeclaration
from semantic_release.version.declarations.toml import TomlVersionDeclaration
from semantic_release.version.translator import VersionTranslator

NonEmptyString = Annotated[str, Field(..., min_length=1)]


class HvcsClient(str, Enum):
    BITBUCKET = "bitbucket"
    GITHUB = "github"
    GITLAB = "gitlab"
    GITEA = "gitea"


_known_commit_parsers: dict[str, type[CommitParser[Any, Any]]] = {
    "angular": AngularCommitParser,
    "conventional": ConventionalCommitParser,
    "conventional-monorepo": ConventionalCommitMonorepoParser,
    "emoji": EmojiCommitParser,
    "scipy": ScipyCommitParser,
    "tag": TagCommitParser,
}


_known_hvcs: Dict[HvcsClient, Type[hvcs.HvcsBase]] = {
    HvcsClient.BITBUCKET: hvcs.Bitbucket,
    HvcsClient.GITHUB: hvcs.Github,
    HvcsClient.GITLAB: hvcs.Gitlab,
    HvcsClient.GITEA: hvcs.Gitea,
}


class EnvConfigVar(BaseModel):
    env: str
    default: Optional[str] = None
    default_env: Optional[str] = None

    def getvalue(self) -> Optional[str]:
        return os.getenv(self.env, os.getenv(self.default_env or "", self.default))


MaybeFromEnv = Union[EnvConfigVar, str]


class ChangelogOutputFormat(str, Enum):
    """Supported changelog output formats when using the default templates."""

    MARKDOWN = "md"
    RESTRUCTURED_TEXT = "rst"
    NONE = ""


class ChangelogEnvironmentConfig(BaseModel):
    block_start_string: str = "{%"
    block_end_string: str = "%}"
    variable_start_string: str = "{{"
    variable_end_string: str = "}}"
    comment_start_string: str = "{#"
    comment_end_string: str = "#}"
    line_statement_prefix: Optional[str] = None
    line_comment_prefix: Optional[str] = None
    trim_blocks: bool = False
    lstrip_blocks: bool = False
    newline_sequence: Literal["\n", "\r", "\r\n"] = "\n"
    keep_trailing_newline: bool = False
    extensions: Tuple[str, ...] = ()
    autoescape: Union[bool, str] = False


class DefaultChangelogTemplatesConfig(BaseModel):
    changelog_file: str = "CHANGELOG.md"
    output_format: ChangelogOutputFormat = ChangelogOutputFormat.NONE
    mask_initial_release: bool = True

    @model_validator(mode="after")
    def interpret_output_format(self) -> Self:
        # Set the output format value when no user input is given
        if self.output_format == ChangelogOutputFormat.NONE:
            try:
                # Note: If the user gave no extension, force '.' so enumeration fails,
                # and defaults to markdown
                # Otherwise normal files with extensions will just look for the extension support
                self.output_format = ChangelogOutputFormat(
                    Path(self.changelog_file).suffix.lstrip(".") or "."
                )
            except ValueError:
                self.output_format = ChangelogOutputFormat.MARKDOWN

        return self


class ChangelogConfig(BaseModel):
    # TODO: BREAKING CHANGE v11, move to DefaultChangelogTemplatesConfig
    changelog_file: str = ""
    """Deprecated! Moved to 'default_templates.changelog_file'"""

    default_templates: DefaultChangelogTemplatesConfig = (
        DefaultChangelogTemplatesConfig(output_format=ChangelogOutputFormat.NONE)
    )
    environment: ChangelogEnvironmentConfig = ChangelogEnvironmentConfig()
    exclude_commit_patterns: Tuple[str, ...] = ()
    mode: ChangelogMode = ChangelogMode.UPDATE
    insertion_flag: str = ""
    template_dir: str = "templates"

    @field_validator("exclude_commit_patterns", mode="after")
    @classmethod
    def validate_match(cls, patterns: Tuple[str, ...]) -> Tuple[str, ...]:
        curr_index = 0
        try:
            for i, pattern in enumerate(patterns):
                curr_index = i
                regexp(pattern)
        except RegExpError as err:
            raise ValueError(
                f"exclude_commit_patterns[{curr_index}]: Invalid regular expression"
            ) from err
        return patterns

    @field_validator("changelog_file", mode="after")
    @classmethod
    def changelog_file_deprecation_warning(cls, val: str) -> str:
        logger.warning(
            str.join(
                " ",
                [
                    "The 'changelog.changelog_file' configuration option is moving to 'changelog.default_templates.changelog_file'.",
                    "Please update your configuration as the compatibility will break in v10.",
                ],
            )
        )
        return val

    @model_validator(mode="after")
    def move_changelog_file(self) -> Self:
        # TODO: Remove this method in v11
        if not self.changelog_file:
            return self

        if self.changelog_file == self.default_templates.changelog_file:
            return self

        # Re-evaluate now that we are passing the changelog_file option down to default_templates
        # and only reset the output_format if it was not already set by the user
        self.default_templates = DefaultChangelogTemplatesConfig.model_validate(
            {
                **self.default_templates.model_dump(),
                "changelog_file": self.changelog_file,
                "output_format": (
                    self.default_templates.output_format
                    if self.default_templates.output_format
                    != ChangelogOutputFormat.MARKDOWN
                    else ChangelogOutputFormat.NONE
                ),
            }
        )

        return self

    @model_validator(mode="after")
    def load_default_insertion_flag_on_missing(self) -> Self:
        # Set the insertion flag value when no user input is given
        if not self.insertion_flag:
            defaults = {
                ChangelogOutputFormat.MARKDOWN: "<!-- version list -->",
                ChangelogOutputFormat.RESTRUCTURED_TEXT: f"..{os.linesep}    version list",
            }
            try:
                self.insertion_flag = defaults[self.default_templates.output_format]
            except KeyError as err:
                raise ValueError(
                    "changelog.default_templates.output_format cannot be None"
                ) from err

        return self


class BranchConfig(BaseModel):
    match: str = "(main|master)"
    prerelease_token: str = "rc"  # noqa: S105
    prerelease: bool = False

    @field_validator("match", mode="after")
    @classmethod
    def validate_match(cls, match: str) -> str:
        # Allow the special case of a plain wildcard although it's not a valid regex
        if match == "*":
            return ".*"

        try:
            regexp(match)
        except RegExpError as err:
            raise ValueError(f"Invalid regex {match!r}") from err
        return match


class RemoteConfig(BaseModel):
    name: str = "origin"
    token: Optional[str] = None
    url: Optional[str] = None
    type: HvcsClient = HvcsClient.GITHUB
    domain: Optional[str] = None
    api_domain: Optional[str] = None
    ignore_token_for_push: bool = False
    insecure: bool = False

    @field_validator("url", "domain", "api_domain", "token", mode="before")
    @classmethod
    def resolve_env_vars(cls, val: Any) -> str | None:
        ret_val = (
            val
            if not isinstance(val, dict)
            else (EnvConfigVar.model_validate(val).getvalue())
        )
        return ret_val or None

    @model_validator(mode="after")
    def set_default_token(self) -> Self:
        # Set the default token name for the given VCS when no user input is given
        if self.token:
            return self
        if self.type not in _known_hvcs:
            return self
        if env_token := self._get_default_token():
            self.token = env_token
        return self

    def _get_default_token(self) -> str | None:
        hvcs_client_class = _known_hvcs[self.type]

        default_token_name = (
            getattr(hvcs_client_class, "DEFAULT_ENV_TOKEN_NAME")  # noqa: B009
            if hasattr(hvcs_client_class, "DEFAULT_ENV_TOKEN_NAME")
            else ""
        )

        return (
            EnvConfigVar(env=default_token_name).getvalue()
            if default_token_name
            else None
        )

    @model_validator(mode="after")
    def check_url_scheme(self) -> Self:
        if self.url and isinstance(self.url, str):
            self.check_insecure_flag(self.url, "url")

        if self.domain and isinstance(self.domain, str):
            self.check_insecure_flag(self.domain, "domain")

        if self.api_domain and isinstance(self.api_domain, str):
            self.check_insecure_flag(self.api_domain, "api_domain")

        return self

    def check_insecure_flag(self, url_str: str, field_name: str) -> None:
        if not url_str:
            return

        scheme = parse_url(url_str).scheme
        if scheme == "http" and not self.insecure:
            raise ValueError(
                str.join(
                    "\n",
                    [
                        "Insecure 'HTTP' URL detected and disabled by default.",
                        "Set the 'insecure' flag to 'True' to enable insecure connections.",
                    ],
                )
            )

        if scheme == "https" and self.insecure:
            logger.warning(
                str.join(
                    "\n",
                    [
                        f"'{field_name}' starts with 'https://' but the 'insecure' flag is set.",
                        "This flag is only necessary for 'http://' URLs.",
                    ],
                )
            )


class PublishConfig(BaseModel):
    dist_glob_patterns: Tuple[str, ...] = ("dist/*",)
    upload_to_vcs_release: bool = True


class RawConfig(BaseModel):
    assets: List[str] = []
    branches: Dict[str, BranchConfig] = {"main": BranchConfig()}
    build_command: Optional[str] = None
    build_command_env: List[str] = []
    changelog: ChangelogConfig = ChangelogConfig()
    commit_author: MaybeFromEnv = EnvConfigVar(
        env="GIT_COMMIT_AUTHOR", default=DEFAULT_COMMIT_AUTHOR
    )
    commit_message: str = COMMIT_MESSAGE
    commit_parser: NonEmptyString = "conventional"
    # It's up to the parser_options() method to validate these
    commit_parser_options: Dict[str, Any] = {}
    logging_use_named_masks: bool = False
    major_on_zero: bool = True
    allow_zero_version: bool = False
    repo_dir: Annotated[Path, Field(validate_default=True)] = Path(".")
    remote: RemoteConfig = RemoteConfig()
    no_git_verify: bool = False
    tag_format: str = "v{version}"
    add_partial_tags: bool = False
    publish: PublishConfig = PublishConfig()
    version_toml: Optional[Tuple[str, ...]] = None
    version_variables: Optional[Tuple[str, ...]] = None

    @field_validator("repo_dir", mode="before")
    @classmethod
    def convert_str_to_path(cls, value: Any) -> Path:
        if not isinstance(value, (str, Path)):
            raise TypeError(f"Invalid type: {type(value)}, expected str or Path.")
        return Path(value)

    @field_validator("repo_dir", mode="after")
    @classmethod
    def verify_git_repo_dir(cls, dir_path: Path) -> Path:
        try:
            # Check for repository & walk up parent directories
            with Repo(str(dir_path), search_parent_directories=True) as git_repo:
                found_path = (
                    Path(git_repo.working_tree_dir or git_repo.working_dir)
                    .expanduser()
                    .absolute()
                )

        except InvalidGitRepositoryError as err:
            raise InvalidGitRepositoryError("No valid git repository found!") from err

        if dir_path.absolute() != found_path:
            logging.warning(
                "Found .git/ in higher parent directory rather than provided in configuration."
            )

        return found_path.resolve()

    @field_validator("commit_parser", mode="after")
    @classmethod
    def tag_commit_parser_deprecation_warning(cls, val: str) -> str:
        if val == "tag":
            logger.warning(
                str.join(
                    " ",
                    [
                        "The legacy 'tag' parser is deprecated and will be removed in v11.",
                        "Recommend swapping to our emoji parser (higher-compatibility)",
                        "or switch to another supported parser.",
                    ],
                )
            )
        return val

    @field_validator("commit_parser", mode="after")
    @classmethod
    def angular_commit_parser_deprecation_warning(cls, val: str) -> str:
        if val == "angular":
            logger.warning(
                str.join(
                    " ",
                    [
                        "The 'angular' parser is deprecated and will be removed in v11.",
                        "The Angular parser is being renamed to the conventional commit parser,",
                        "which is selected by switching the 'commit_parser' value to 'conventional'.",
                    ],
                )
            )
        return val

    @field_validator("build_command_env", mode="after")
    @classmethod
    def remove_whitespace(cls, val: list[str]) -> list[str]:
        return [entry.strip() for entry in val]

    @model_validator(mode="after")
    def set_default_opts(self) -> Self:
        # Set the default parser options for the given commit parser when no user input is given
        if not self.commit_parser_options and self.commit_parser:
            parser_opts_type = None
            # If the commit parser is a known one, pull the default options object from it
            if self.commit_parser in _known_commit_parsers:
                # TODO: BREAKING CHANGE v11
                # parser_opts_type = (
                #     _known_commit_parsers[self.commit_parser]
                #     .get_default_options()
                #     .__class__
                # )
                parser_opts_type = _known_commit_parsers[
                    self.commit_parser
                ].parser_options
            else:
                try:
                    # if its a custom parser, try to import it and pull the default options object type
                    custom_class = dynamic_import(self.commit_parser)
                    # TODO: BREAKING CHANGE v11
                    # parser_opts_type = custom_class.get_default_options().__class__
                    if hasattr(custom_class, "parser_options"):
                        parser_opts_type = custom_class.parser_options

                except ModuleNotFoundError as err:
                    raise ParserLoadError(
                        str.join(
                            "\n",
                            [
                                str(err),
                                "Unable to import your custom parser! Check your configuration!",
                            ],
                        )
                    ) from err

                except AttributeError as err:
                    raise ParserLoadError(
                        str.join(
                            "\n",
                            [
                                str(err),
                                "Unable to find your custom parser class inside the given module.",
                                "Check your configuration!",
                            ],
                        )
                    ) from err

            # from either the custom opts class or the known parser opts class, create an instance
            if callable(parser_opts_type):
                opts_obj: Any = parser_opts_type()
                # if the opts object is a dataclass, wrap it in a RootModel so it can be transformed to a Mapping
                opts_obj = (
                    opts_obj if not is_dataclass(opts_obj) else RootModel(opts_obj)
                )
                # Must be a mapping, so if it's a BaseModel, dump the model to a dict
                self.commit_parser_options = (
                    opts_obj.model_dump()
                    if isinstance(opts_obj, (BaseModel, RootModel))
                    else opts_obj
                )
                if not isinstance(self.commit_parser_options, Mapping):
                    raise ValidationError(
                        f"Invalid parser options: {opts_obj}. Must be a mapping."
                    )

        return self


@dataclass
class GlobalCommandLineOptions:
    """
    A dataclass to hold all the command line options that
    should be set in the RuntimeContext
    """

    noop: bool = False
    verbosity: int = 0
    config_file: str = DEFAULT_CONFIG_FILE
    strict: bool = False


######
# RuntimeContext
######
# This is what we want to attach to `click.Context.obj`
# There are currently no defaults here - this is on purpose,
# the defaults should be specified and handled by `RawConfig`.
# When this is constructed we should know exactly what the user
# wants
def _recursive_getattr(obj: Any, path: str) -> Any:
    """
    Used to find nested parts of RuntimeContext which
    might contain sensitive data. Returns None if an attribute
    is missing
    """
    out = obj
    for part in path.split("."):
        out = getattr(out, part, None)
    return out


@dataclass
class RuntimeContext:
    _mask_attrs_: ClassVar[List[str]] = ["hvcs_client.token"]

    project_metadata: dict[str, Any]
    repo_dir: Path
    commit_parser: CommitParser[ParseResult, ParserOptions]
    version_translator: VersionTranslator
    major_on_zero: bool
    allow_zero_version: bool
    prerelease: bool
    no_git_verify: bool
    assets: List[str]
    commit_author: Actor
    commit_message: str
    changelog_excluded_commit_patterns: Tuple[Pattern[str], ...]
    version_declarations: Tuple[IVersionReplacer, ...]
    hvcs_client: hvcs.HvcsBase
    changelog_insertion_flag: str
    changelog_mask_initial_release: bool
    changelog_mode: ChangelogMode
    changelog_file: Path
    changelog_style: str
    changelog_output_format: ChangelogOutputFormat
    ignore_token_for_push: bool
    template_environment: Environment
    template_dir: Path
    build_command: Optional[str]
    build_command_env: dict[str, str]
    dist_glob_patterns: Tuple[str, ...]
    upload_to_vcs_release: bool
    global_cli_options: GlobalCommandLineOptions
    # This way the filter can be passed around if needed, so that another function
    # can accept the filter as an argument and call
    masker: MaskingFilter

    @staticmethod
    def resolve_from_env(param: Optional[MaybeFromEnv]) -> Optional[str]:
        if isinstance(param, EnvConfigVar):
            return param.getvalue()
        return param

    @staticmethod
    def select_branch_options(
        choices: Dict[str, BranchConfig], active_branch: str
    ) -> BranchConfig:
        for group, options in choices.items():
            if regexp(options.match).match(active_branch):
                logger.info(
                    "Using group %r options, as %r matches %r",
                    group,
                    options.match,
                    active_branch,
                )
                return options
            logger.debug(
                "Rejecting group %r as %r doesn't match %r",
                group,
                options.match,
                active_branch,
            )

        raise NotAReleaseBranch(
            f"branch {active_branch!r} isn't in any release groups; "
            "no release will be made"
        )

    def apply_log_masking(self, masker: MaskingFilter) -> MaskingFilter:
        for attr in self._mask_attrs_:
            masker.add_mask_for(str(_recursive_getattr(self, attr)), f"context.{attr}")
            masker.add_mask_for(repr(_recursive_getattr(self, attr)), f"context.{attr}")
        return masker

    @classmethod
    def from_raw_config(  # noqa: C901
        cls, raw: RawConfig, global_cli_options: GlobalCommandLineOptions
    ) -> RuntimeContext:
        ##
        # credentials masking for logging
        masker = MaskingFilter(_use_named_masks=raw.logging_use_named_masks)

        # TODO: move to config if we change how the generated config is constructed
        # Retrieve project metadata from pyproject.toml
        project_metadata: dict[str, str] = {}
        curr_dir = Path.cwd().resolve()
        allowed_directories = [
            dir_path
            for dir_path in [curr_dir, *curr_dir.parents]
            if str(raw.repo_dir) in str(dir_path)
        ]
        for allowed_dir in allowed_directories:
            if (proj_toml := allowed_dir.joinpath("pyproject.toml")).exists():
                config_toml = tomlkit.parse(proj_toml.read_text())
                project_metadata = config_toml.unwrap().get("project", project_metadata)
                break

        # Retrieve details from repository
        with Repo(str(raw.repo_dir)) as git_repo:
            try:
                # Get the remote url by calling out to `git remote get-url`. This returns
                # the expanded url, taking into account any insteadOf directives
                # in the git configuration.
                remote_url = raw.remote.url or git_repo.git.remote(
                    "get-url", raw.remote.name
                )
                active_branch = git_repo.active_branch.name
            except ValueError as err:
                raise MissingGitRemote(
                    f"Unable to locate remote named '{raw.remote.name}'."
                ) from err
            except TypeError as err:
                raise DetachedHeadGitError(
                    "Detached HEAD state cannot match any release groups; "
                    "no release will be made"
                ) from err

        # branch-specific configuration
        branch_config = cls.select_branch_options(raw.branches, active_branch)

        # commit_parser
        try:
            commit_parser_cls = (
                _known_commit_parsers[raw.commit_parser]
                if raw.commit_parser in _known_commit_parsers
                else dynamic_import(raw.commit_parser)
            )
        except ValueError as err:
            raise ParserLoadError(
                str.join(
                    "\n",
                    [
                        f"Unrecognized commit parser value: {raw.commit_parser!r}.",
                        str(err),
                        "Unable to load the given parser! Check your configuration!",
                    ],
                )
            ) from err
        except ModuleNotFoundError as err:
            raise ParserLoadError(
                str.join(
                    "\n",
                    [
                        str(err),
                        "Unable to import your custom parser! Check your configuration!",
                    ],
                )
            ) from err
        except AttributeError as err:
            raise ParserLoadError(
                str.join(
                    "\n",
                    [
                        str(err),
                        "Unable to find the parser class inside the given module",
                    ],
                )
            ) from err

        commit_parser_opts_class = commit_parser_cls.parser_options
        # TODO: Breaking change v11
        # commit_parser_opts_class = commit_parser_cls.get_default_options().__class__
        try:
            commit_parser = commit_parser_cls(
                options=commit_parser_opts_class(**raw.commit_parser_options)
            )
        except TypeError as err:
            raise ParserLoadError(
                str.join("\n", [str(err), f"Failed to initialize {raw.commit_parser}"])
            ) from err

        # We always exclude PSR's own release commits from the Changelog
        # when parsing commits
        psr_release_commit_regex = regexp(
            reduce(
                lambda regex_str, pattern: str(regex_str).replace(*pattern),
                (
                    # replace the version holder with a regex pattern to match various versions
                    (regex_escape("{version}"), r"(?P<version>\d+\.\d+\.\d+\S*)"),
                    # TODO: add any other placeholders here
                ),
                # We use re.escape to ensure that the commit message is treated as a literal
                regex_escape(raw.commit_message.strip()),
            )
        )
        changelog_excluded_commit_patterns = (
            psr_release_commit_regex,
            *(regexp(pattern) for pattern in raw.changelog.exclude_commit_patterns),
        )

        _commit_author_str = cls.resolve_from_env(raw.commit_author) or ""
        _commit_author_valid = Actor.name_email_regex.match(_commit_author_str)
        if not _commit_author_valid:
            raise ValueError(
                f"Invalid git author: {_commit_author_str} "
                f"should match {Actor.name_email_regex}"
            )

        commit_author = Actor(*_commit_author_valid.groups())

        version_declarations: list[IVersionReplacer] = []

        try:
            version_declarations.extend(
                TomlVersionDeclaration.from_string_definition(definition)
                for definition in iter(raw.version_toml or ())
            )
        except ValueError as err:
            raise InvalidConfiguration(
                str.join(
                    "\n",
                    [
                        "Invalid 'version_toml' configuration",
                        str(err),
                    ],
                )
            ) from err

        try:
            version_declarations.extend(
                PatternVersionDeclaration.from_string_definition(
                    definition, raw.tag_format
                )
                for definition in iter(raw.version_variables or ())
            )
        except ValueError as err:
            raise InvalidConfiguration(
                str.join(
                    "\n",
                    [
                        "Invalid 'version_variables' configuration",
                        str(err),
                    ],
                )
            ) from err

        # Provide warnings if the token is missing
        if not raw.remote.token:
            logger.debug("hvcs token is not set")

            if not raw.remote.ignore_token_for_push:
                logger.warning("Token value is missing!")

        # hvcs_client
        hvcs_client_cls = _known_hvcs[raw.remote.type]
        hvcs_client = hvcs_client_cls(
            remote_url=remote_url,
            hvcs_domain=raw.remote.domain,
            hvcs_api_domain=raw.remote.api_domain,
            token=raw.remote.token,
            allow_insecure=raw.remote.insecure,
        )

        # changelog_file
        # Must use absolute after resolve because windows does not resolve if the path does not exist
        # which means it returns a relative path. So we force absolute to ensure path is complete
        # for the next check of path matching
        changelog_file = (
            Path(raw.changelog.default_templates.changelog_file)
            .expanduser()
            .resolve()
            .absolute()
        )

        # Prevent path traversal attacks
        if raw.repo_dir not in changelog_file.parents:
            raise InvalidConfiguration(
                "Changelog file destination must be inside of the repository directory."
            )

        # Must use absolute after resolve because windows does not resolve if the path does not exist
        # which means it returns a relative path. So we force absolute to ensure path is complete
        # for the next check of path matching
        template_dir = (
            Path(raw.changelog.template_dir).expanduser().resolve().absolute()
        )

        # Prevent path traversal attacks
        if raw.repo_dir not in template_dir.parents:
            raise InvalidConfiguration(
                "Template directory must be inside of the repository directory."
            )

        template_environment = environment(
            template_dir=template_dir,
            **raw.changelog.environment.model_dump(),
        )

        # version_translator
        version_translator = VersionTranslator(
            tag_format=raw.tag_format,
            prerelease_token=branch_config.prerelease_token,
            add_partial_tags=raw.add_partial_tags,
        )

        build_cmd_env = {}

        for i, env_var_def in enumerate(raw.build_command_env):
            # creative hack to handle, missing =, but also = that then can be unpacked
            # as the resulting parts array can be either 2 or 3 in length. it becomes 3
            # with our forced empty value at the end which can be dropped
            parts = [*env_var_def.split("=", 1), ""]
            # removes any odd spacing around =, and extracts name=value
            name, env_val = (part.strip() for part in parts[:2])

            if not name:
                # Skip when invalid format (ex. starting with = and no name)
                logging.warning(
                    "Skipping invalid build_command_env[%s] definition",
                    i,
                )
                continue

            if not env_val and env_var_def[-1] != "=":
                # avoid the edge case that user wants to define a value as empty
                # and don't autoresolve it
                env_val = os.getenv(name, "")

            build_cmd_env[name] = env_val

        # TODO: better support for custom parsers that actually just extend defaults
        #
        # Here we just assume the desired changelog style matches the parser name
        # as we provide templates specific to each parser type. Unfortunately if the user has
        # provided a custom parser, it would be up to the user to provide custom templates
        # but we just assume the base template is conventional
        # changelog_style = (
        #     raw.commit_parser
        #     if raw.commit_parser in _known_commit_parsers
        #     else "conventional"
        # )

        self = cls(
            project_metadata=project_metadata,
            repo_dir=raw.repo_dir,
            commit_parser=commit_parser,
            version_translator=version_translator,
            major_on_zero=raw.major_on_zero,
            allow_zero_version=raw.allow_zero_version,
            build_command=raw.build_command,
            build_command_env=build_cmd_env,
            version_declarations=tuple(version_declarations),
            hvcs_client=hvcs_client,
            changelog_file=changelog_file,
            changelog_mode=raw.changelog.mode,
            changelog_mask_initial_release=raw.changelog.default_templates.mask_initial_release,
            changelog_insertion_flag=raw.changelog.insertion_flag,
            assets=raw.assets,
            commit_author=commit_author,
            commit_message=raw.commit_message,
            changelog_excluded_commit_patterns=changelog_excluded_commit_patterns,
            # TODO: change when we have other styles per parser
            # changelog_style=changelog_style,
            changelog_style="conventional",
            changelog_output_format=raw.changelog.default_templates.output_format,
            prerelease=branch_config.prerelease,
            ignore_token_for_push=raw.remote.ignore_token_for_push,
            template_dir=template_dir,
            template_environment=template_environment,
            dist_glob_patterns=raw.publish.dist_glob_patterns,
            upload_to_vcs_release=raw.publish.upload_to_vcs_release,
            global_cli_options=global_cli_options,
            masker=masker,
            no_git_verify=raw.no_git_verify,
        )
        # credential masker
        self.apply_log_masking(self.masker)

        return self



from __future__ import annotations

from datetime import datetime, timedelta, timezone
from os import remove as delete_file
from textwrap import dedent
from typing import TYPE_CHECKING

import pytest
from freezegun import freeze_time
from pytest_lazy_fixtures.lazy_fixture import lf as lazy_fixture

from semantic_release.changelog.context import ChangelogMode

from tests.const import (
    MAIN_PROG_NAME,
    VERSION_SUBCMD,
)
from tests.e2e.conftest import (
    get_sanitized_md_changelog_content,
    get_sanitized_rst_changelog_content,
)
from tests.fixtures.example_project import (
    changelog_md_file,
    changelog_rst_file,
)
from tests.fixtures.repos import (
    repo_w_trunk_only_conventional_commits,
)
from tests.util import (
    assert_successful_exit_code,
)

if TYPE_CHECKING:
    from pathlib import Path
    from typing import TypedDict

    from tests.conftest import GetCachedRepoDataFn, GetStableDateNowFn, RunCliFn
    from tests.e2e.conftest import GetSanitizedChangelogContentFn
    from tests.fixtures.example_project import UpdatePyprojectTomlFn
    from tests.fixtures.git_repo import (
        BuildRepoFromDefinitionFn,
        BuiltRepoResult,
        CommitDef,
        GetCfgValueFromDefFn,
        GetVersionsFromRepoBuildDefFn,
        SplitRepoActionsByReleaseTagsFn,
    )

    class Commit2Section(TypedDict):
        conventional: Commit2SectionCommit
        emoji: Commit2SectionCommit
        scipy: Commit2SectionCommit

    class Commit2SectionCommit(TypedDict):
        commit: CommitDef
        section: str


@pytest.mark.parametrize(
    str.join(
        ", ",
        [
            "custom_commit_message",
            "changelog_mode",
            "changelog_file",
            "get_sanitized_changelog_content",
            "repo_result",
            "repo_fixture_name",
        ],
    ),
    [
        pytest.param(
            custom_commit_message,
            changelog_mode,
            lazy_fixture(changelog_file),
            lazy_fixture(cl_sanitizer),
            lazy_fixture(repo_fixture_name),
            repo_fixture_name,
            marks=pytest.mark.comprehensive,
        )
        for changelog_mode in [ChangelogMode.INIT, ChangelogMode.UPDATE]
        for changelog_file, cl_sanitizer in [
            (
                changelog_md_file.__name__,
                get_sanitized_md_changelog_content.__name__,
            ),
            (
                changelog_rst_file.__name__,
                get_sanitized_rst_changelog_content.__name__,
            ),
        ]
        for repo_fixture_name, custom_commit_message in [
            *[
                (
                    # Repos: Must have at least 2 releases
                    repo_w_trunk_only_conventional_commits.__name__,
                    commit_msg,
                )
                for commit_msg in [
                    dedent(
                        # Conventional compliant prefix with skip-ci indicator
                        """\
                        chore(release): v{version} [skip ci]

                        Automatically generated by python-semantic-release.
                        """
                    ),
                ]
            ],
        ]
    ],
)
def test_version_changelog_content_custom_commit_message_excluded_automatically(
    repo_result: BuiltRepoResult,
    get_versions_from_repo_build_def: GetVersionsFromRepoBuildDefFn,
    get_cfg_value_from_def: GetCfgValueFromDefFn,
    split_repo_actions_by_release_tags: SplitRepoActionsByReleaseTagsFn,
    build_repo_from_definition: BuildRepoFromDefinitionFn,
    run_cli: RunCliFn,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    changelog_file: Path,
    changelog_mode: ChangelogMode,
    custom_commit_message: str,
    repo_fixture_name: str,
    stable_now_date: GetStableDateNowFn,
    example_project_dir: Path,
    get_sanitized_changelog_content: GetSanitizedChangelogContentFn,
    get_cached_repo_data: GetCachedRepoDataFn,
):
    """
    Given a repo with a custom release commit message
    When the version subcommand is invoked with the changelog flag
    Then the resulting changelog content should not include the
    custom commit message

    It should work regardless of changelog mode and changelog file type
    """
    expected_changelog_content = get_sanitized_changelog_content(
        repo_dir=example_project_dir,
        remove_insertion_flag=bool(changelog_mode == ChangelogMode.INIT),
    )

    repo = repo_result["repo"]
    repo_def = repo_result["definition"]
    tag_format_str: str = get_cfg_value_from_def(repo_def, "tag_format_str")  # type: ignore[assignment]
    all_versions = get_versions_from_repo_build_def(repo_def)
    latest_version = all_versions[-1]
    previous_tag = tag_format_str.format(version=all_versions[-2])

    # split repo actions by release actions
    releasetags_2_steps = split_repo_actions_by_release_tags(repo_def)

    # Reverse release to make the previous version again with the new commit message
    repo.git.tag("-d", latest_version.as_tag())
    repo.git.reset("--hard", f"{previous_tag}~1")
    repo.git.tag("-d", previous_tag)

    # Set the project configurations
    update_pyproject_toml("tool.semantic_release.changelog.mode", changelog_mode.value)
    update_pyproject_toml(
        "tool.semantic_release.changelog.default_templates.changelog_file",
        str(changelog_file.name),
    )
    update_pyproject_toml(
        "tool.semantic_release.commit_message",
        custom_commit_message,
    )

    if not (repo_build_data := get_cached_repo_data(repo_fixture_name)):
        pytest.fail("Repo build date not found in cache")

    repo_build_datetime = datetime.strptime(repo_build_data["build_date"], "%Y-%m-%d")
    now_datetime = stable_now_date().replace(
        year=repo_build_datetime.year,
        month=repo_build_datetime.month,
        day=repo_build_datetime.day,
    )

    if changelog_mode == ChangelogMode.UPDATE and len(all_versions) == 2:
        # When in update mode, and at the very first release, its better the
        # changelog file does not exist as we have an non-conformative example changelog
        # in the base example project
        delete_file(example_project_dir / changelog_file)

    cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, "--no-push", "--changelog"]

    # Act: make the first release again
    with freeze_time(now_datetime.astimezone(timezone.utc)):
        result = run_cli(cli_cmd[1:])
        assert_successful_exit_code(result, cli_cmd)

    # Act: apply commits for change of version
    # stop before the release step
    steps_for_next_release = releasetags_2_steps[latest_version][:-1]
    build_repo_from_definition(
        dest_dir=example_project_dir,
        repo_construction_steps=steps_for_next_release,
    )

    # Act: make the second release again
    with freeze_time(now_datetime.astimezone(timezone.utc) + timedelta(minutes=1)):
        result = run_cli(cli_cmd[1:])

    actual_content = get_sanitized_changelog_content(
        repo_dir=example_project_dir,
        remove_insertion_flag=bool(changelog_mode == ChangelogMode.INIT),
    )

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)
    assert expected_changelog_content == actual_content



"""Common functionality and interface for interacting with Git remote VCS"""

from __future__ import annotations

import warnings
from abc import ABCMeta, abstractmethod
from functools import lru_cache
from typing import TYPE_CHECKING

from semantic_release.helpers import parse_git_url

if TYPE_CHECKING:  # pragma: no cover
    from typing import Any, Callable


class HvcsBase(metaclass=ABCMeta):
    """
    Interface for subclasses interacting with a remote vcs environment

    Methods generally have a base implementation are implemented here but
    likely just provide a not-supported message but return gracefully

    This class cannot be instantated directly but must be inherited from
    and implement the designated abstract methods.
    """

    def __init__(self, remote_url: str, *args: Any, **kwargs: Any) -> None:
        self._remote_url = remote_url if parse_git_url(remote_url) else ""
        self._name: str | None = None
        self._owner: str | None = None

    def _not_supported(self: HvcsBase, method_name: str) -> None:
        warnings.warn(
            f"{method_name} is not supported by {type(self).__qualname__}",
            stacklevel=2,
        )

    @lru_cache(maxsize=1)
    def _get_repository_owner_and_name(self) -> tuple[str, str]:
        """
        Parse the repository's remote url to identify the repository
        owner and name
        """
        parsed_git_url = parse_git_url(self._remote_url)
        return parsed_git_url.namespace, parsed_git_url.repo_name

    @property
    def repo_name(self) -> str:
        if self._name is None:
            _, name = self._get_repository_owner_and_name()
            self._name = name
        return self._name

    @property
    def owner(self) -> str:
        if self._owner is None:
            _owner, _ = self._get_repository_owner_and_name()
            self._owner = _owner
        return self._owner

    @abstractmethod
    def remote_url(self, use_token: bool) -> str:
        """
        Return the remote URL for the repository, including the token for
        authentication if requested by setting the `use_token` parameter to True,
        """
        self._not_supported(self.remote_url.__name__)
        return ""

    @abstractmethod
    def get_changelog_context_filters(self) -> tuple[Callable[..., Any], ...]:
        """
        Return a list of functions that can be used as filters in a Jinja2 template

        ex. filters to convert text to URLs for issues and commits
        """
        self._not_supported(self.get_changelog_context_filters.__name__)
        return ()



from __future__ import annotations

import os
from datetime import timezone
from typing import TYPE_CHECKING

import pytest
from freezegun import freeze_time
from pytest_lazy_fixtures.lazy_fixture import lf as lazy_fixture

from semantic_release.version.version import Version

from tests.const import (
    EXAMPLE_RELEASE_NOTES_TEMPLATE,
    MAIN_PROG_NAME,
    VERSION_SUBCMD,
    RepoActionStep,
)
from tests.fixtures.repos import repo_w_no_tags_conventional_commits
from tests.fixtures.repos.trunk_based_dev.repo_w_no_tags import (
    repo_w_no_tags_emoji_commits,
    repo_w_no_tags_scipy_commits,
)
from tests.util import assert_successful_exit_code, get_release_history_from_context

if TYPE_CHECKING:
    from unittest.mock import MagicMock

    from requests_mock import Mocker

    from tests.conftest import GetStableDateNowFn, RunCliFn
    from tests.e2e.conftest import (
        RetrieveRuntimeContextFn,
    )
    from tests.fixtures.example_project import (
        UpdatePyprojectTomlFn,
        UseReleaseNotesTemplateFn,
    )
    from tests.fixtures.git_repo import (
        BuiltRepoResult,
        GenerateDefaultReleaseNotesFromDefFn,
        GetHvcsClientFromRepoDefFn,
    )


@pytest.mark.parametrize(
    "repo_result, next_release_version",
    [
        (lazy_fixture(repo_w_no_tags_conventional_commits.__name__), "1.0.0"),
    ],
)
def test_custom_release_notes_template(
    repo_result: BuiltRepoResult,
    next_release_version: str,
    run_cli: RunCliFn,
    use_release_notes_template: UseReleaseNotesTemplateFn,
    retrieve_runtime_context: RetrieveRuntimeContextFn,
    mocked_git_fetch: MagicMock,
    mocked_git_push: MagicMock,
    post_mocker: Mocker,
) -> None:
    """Verify the template `.release_notes.md.j2` from `template_dir` is used."""
    release_version = Version.parse(next_release_version)

    # Setup
    use_release_notes_template()
    runtime_context = retrieve_runtime_context(repo_result["repo"])

    # Act
    cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, "--vcs-release"]
    result = run_cli(cli_cmd[1:])

    # Must run this after the action because the release history object should be pulled from the
    # repository after a tag is created
    release_history = get_release_history_from_context(runtime_context)
    release = release_history.released[release_version]

    expected_release_notes = (
        runtime_context.template_environment.from_string(EXAMPLE_RELEASE_NOTES_TEMPLATE)
        .render(release=release)
        .rstrip()
        + os.linesep
    )

    # ensure normalized line endings after render
    expected_release_notes = str.join(
        os.linesep,
        str.split(expected_release_notes.replace("\r", ""), "\n"),
    )

    # Assert
    assert_successful_exit_code(result, cli_cmd)
    assert mocked_git_fetch.call_count == 1  # fetch called to check for remote changes
    assert mocked_git_push.call_count == 2  # 1 for commit, 1 for tag
    assert post_mocker.call_count == 1
    assert post_mocker.last_request is not None

    actual_notes = post_mocker.last_request.json()["body"]
    assert expected_release_notes == actual_notes


@pytest.mark.parametrize(
    "repo_result, license_name, license_setting, mask_initial_release",
    [
        pytest.param(
            lazy_fixture(repo_fixture_name),
            license_name,
            license_setting,
            mask_initial_release,
            marks=pytest.mark.comprehensive,
        )
        for mask_initial_release in [True, False]
        for license_name in ["", "MIT", "GPL-3.0"]
        for license_setting in [
            "project.license-expression",
            "project.license",  # deprecated
            "project.license.text",  # deprecated
        ]
        for repo_fixture_name in [
            repo_w_no_tags_conventional_commits.__name__,
            repo_w_no_tags_emoji_commits.__name__,
            repo_w_no_tags_scipy_commits.__name__,
        ]
    ],
)
def test_default_release_notes_license_statement(
    repo_result: BuiltRepoResult,
    run_cli: RunCliFn,
    license_name: str,
    license_setting: str,
    mask_initial_release: bool,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    mocked_git_fetch: MagicMock,
    mocked_git_push: MagicMock,
    post_mocker: Mocker,
    stable_now_date: GetStableDateNowFn,
    get_hvcs_client_from_repo_def: GetHvcsClientFromRepoDefFn,
    generate_default_release_notes_from_def: GenerateDefaultReleaseNotesFromDefFn,
):
    new_version = "1.0.0"

    # Setup
    now_datetime = stable_now_date()
    repo_def = list(repo_result["definition"])
    repo_def.append(
        {
            "action": RepoActionStep.RELEASE,
            "details": {
                "version": new_version,
                "datetime": now_datetime.isoformat(timespec="seconds"),
            },
        }
    )
    # Setup: Overwrite the default setting (defined in test.const)
    update_pyproject_toml("project.license-expression", None)

    # Setup: set the license for the test
    update_pyproject_toml(license_setting, license_name)

    # Setup: set mask_initial_release value in configuration
    update_pyproject_toml(
        "tool.semantic_release.changelog.default_templates.mask_initial_release",
        mask_initial_release,
    )

    expected_release_notes = generate_default_release_notes_from_def(
        version_actions=repo_def,
        hvcs=get_hvcs_client_from_repo_def(repo_def),
        previous_version=None,
        license_name=license_name,
        mask_initial_release=mask_initial_release,
    )

    # Act
    with freeze_time(now_datetime.astimezone(timezone.utc)):
        cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, "--no-changelog", "--vcs-release"]
        result = run_cli(cli_cmd[1:])

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)
    assert mocked_git_fetch.call_count == 1  # fetch called to check for remote changes
    assert mocked_git_push.call_count == 2  # 1 for commit, 1 for tag
    assert post_mocker.call_count == 1
    assert post_mocker.last_request is not None
    request_body = post_mocker.last_request.json()

    assert "body" in request_body
    actual_notes = request_body["body"]

    assert expected_release_notes == actual_notes



"""E2E tests for upstream verification during version command."""

from __future__ import annotations

import contextlib
from pathlib import PureWindowsPath
from typing import TYPE_CHECKING, cast

import pytest
from git import Repo
from pytest_lazy_fixtures.lazy_fixture import lf as lazy_fixture

from semantic_release.hvcs.github import Github

from tests.const import MAIN_PROG_NAME, VERSION_SUBCMD
from tests.fixtures.example_project import change_to_ex_proj_dir
from tests.fixtures.repos.trunk_based_dev import (
    repo_w_trunk_only_conventional_commits,
)
from tests.fixtures.repos.trunk_based_dev.repo_w_tags import (
    build_trunk_only_repo_w_tags,
)
from tests.util import (
    add_text_to_file,
    assert_exit_code,
    assert_successful_exit_code,
    temporary_working_directory,
)

if TYPE_CHECKING:
    from pathlib import Path

    from requests_mock import Mocker

    from tests.conftest import RunCliFn
    from tests.fixtures.example_project import ExProjectDir, UpdatePyprojectTomlFn
    from tests.fixtures.git_repo import (
        BuildSpecificRepoFn,
        CommitConvention,
        GetCfgValueFromDefFn,
        GetGitRepo4DirFn,
        GetVersionsFromRepoBuildDefFn,
    )


@pytest.mark.parametrize(
    "repo_fixture_name, build_repo_fn",
    [
        (
            repo_fixture_name,
            lazy_fixture(build_repo_fn_name),
        )
        for repo_fixture_name, build_repo_fn_name in [
            (
                repo_w_trunk_only_conventional_commits.__name__,
                build_trunk_only_repo_w_tags.__name__,
            ),
        ]
    ],
)
@pytest.mark.usefixtures(change_to_ex_proj_dir.__name__)
def test_version_upstream_check_success_no_changes(
    repo_fixture_name: str,
    run_cli: RunCliFn,
    build_repo_fn: BuildSpecificRepoFn,
    example_project_dir: ExProjectDir,
    git_repo_for_directory: GetGitRepo4DirFn,
    post_mocker: Mocker,
    get_cfg_value_from_def: GetCfgValueFromDefFn,
    get_versions_from_repo_build_def: GetVersionsFromRepoBuildDefFn,
    pyproject_toml_file: Path,
    update_pyproject_toml: UpdatePyprojectTomlFn,
):
    remote_name = "origin"
    # Create a bare remote (simulating origin)
    local_origin = Repo.init(str(example_project_dir / "local_origin"), bare=True)

    # build target repo into a temporary directory
    target_repo_dir = example_project_dir / repo_fixture_name
    commit_type: CommitConvention = (
        repo_fixture_name.split("commits", 1)[0].split("_")[-2]  # type: ignore[assignment]
    )
    target_repo_definition = build_repo_fn(
        repo_name=repo_fixture_name,
        commit_type=commit_type,
        dest_dir=target_repo_dir,
    )
    target_git_repo = git_repo_for_directory(target_repo_dir)

    # Configure the source repo to use the bare remote (removing any existing 'origin')
    with contextlib.suppress(AttributeError):
        target_git_repo.delete_remote(target_git_repo.remotes[remote_name])

    target_git_repo.create_remote(remote_name, str(local_origin.working_dir))

    # Remove last release before pushing to upstream
    tag_format_str = cast(
        "str", get_cfg_value_from_def(target_repo_definition, "tag_format_str")
    )
    latest_tag = tag_format_str.format(
        version=get_versions_from_repo_build_def(target_repo_definition)[-1]
    )
    target_git_repo.git.tag("-d", latest_tag)
    target_git_repo.git.reset("--hard", "HEAD~1")

    # TODO: when available, switch this to use hvcs=none or similar config to avoid token use for push
    update_pyproject_toml(
        "tool.semantic_release.remote.ignore_token_for_push",
        True,
        target_repo_dir / pyproject_toml_file,
    )
    target_git_repo.git.commit(amend=True, no_edit=True, all=True)

    # push the current state to establish the remote (cannot push tags and branches at the same time)
    target_git_repo.git.push(remote_name, all=True)  # all branches
    target_git_repo.git.push(remote_name, tags=True)  # all tags

    # ensure bare remote HEAD points to the active branch so clones can checkout
    local_origin.git.symbolic_ref(
        "HEAD", f"refs/heads/{target_git_repo.active_branch.name}"
    )

    # current remote tags
    remote_origin_tags_before = {tag.name for tag in local_origin.tags}

    # Clone the repo to simulate a local workspace
    test_repo = Repo.clone_from(
        f"file://{local_origin.working_dir}",
        str(example_project_dir / "repo_clone"),
        no_local=True,
    )
    with test_repo.config_writer("repository") as config:
        config.set_value("core", "hookspath", "")
        config.set_value("commit", "gpgsign", False)
        config.set_value("tag", "gpgsign", False)

    current_head_sha = test_repo.head.commit.hexsha

    # Act: run PSR on the cloned repo - it should verify upstream and succeed
    with temporary_working_directory(str(test_repo.working_dir)):
        cli_cmd = [MAIN_PROG_NAME, "--strict", VERSION_SUBCMD]
        result = run_cli(cli_cmd[1:], env={Github.DEFAULT_ENV_TOKEN_NAME: "1234"})

    remote_origin_tags_after = {tag.name for tag in local_origin.tags}

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)

    # Verify release occurred as expected
    with test_repo:
        assert latest_tag in test_repo.tags, "Expected release tag to be created"
        assert current_head_sha in [
            parent.hexsha for parent in test_repo.head.commit.parents
        ], "Expected new commit to be created on HEAD"
        different_tags = remote_origin_tags_after.difference(remote_origin_tags_before)
        assert latest_tag in different_tags, "Expected new tag to be pushed to remote"

    # Verify VCS release was created
    expected_vcs_url_post = 1
    assert expected_vcs_url_post == post_mocker.call_count  # one vcs release created


@pytest.mark.parametrize(
    "repo_fixture_name, build_repo_fn",
    [
        (
            repo_fixture_name,
            lazy_fixture(build_repo_fn_name),
        )
        for repo_fixture_name, build_repo_fn_name in [
            (
                repo_w_trunk_only_conventional_commits.__name__,
                build_trunk_only_repo_w_tags.__name__,
            ),
        ]
    ],
)
@pytest.mark.usefixtures(change_to_ex_proj_dir.__name__)
def test_version_upstream_check_success_no_changes_untracked_branch(
    repo_fixture_name: str,
    run_cli: RunCliFn,
    build_repo_fn: BuildSpecificRepoFn,
    example_project_dir: ExProjectDir,
    git_repo_for_directory: GetGitRepo4DirFn,
    post_mocker: Mocker,
    get_cfg_value_from_def: GetCfgValueFromDefFn,
    get_versions_from_repo_build_def: GetVersionsFromRepoBuildDefFn,
    pyproject_toml_file: Path,
    update_pyproject_toml: UpdatePyprojectTomlFn,
):
    """Test that PSR succeeds when the upstream branch is untracked but unchanged."""
    remote_name = "origin"
    # Create a bare remote (simulating origin)
    local_origin = Repo.init(str(example_project_dir / "local_origin"), bare=True)

    # build target repo into a temporary directory
    target_repo_dir = example_project_dir / repo_fixture_name
    commit_type: CommitConvention = (
        repo_fixture_name.split("commits", 1)[0].split("_")[-2]  # type: ignore[assignment]
    )
    target_repo_definition = build_repo_fn(
        repo_name=repo_fixture_name,
        commit_type=commit_type,
        dest_dir=target_repo_dir,
    )
    target_git_repo = git_repo_for_directory(target_repo_dir)

    # Configure the source repo to use the bare remote (removing any existing 'origin')
    with contextlib.suppress(AttributeError):
        target_git_repo.delete_remote(target_git_repo.remotes[remote_name])

    target_git_repo.create_remote(remote_name, str(local_origin.working_dir))

    # Remove last release before pushing to upstream
    tag_format_str = cast(
        "str", get_cfg_value_from_def(target_repo_definition, "tag_format_str")
    )
    latest_tag = tag_format_str.format(
        version=get_versions_from_repo_build_def(target_repo_definition)[-1]
    )
    target_git_repo.git.tag("-d", latest_tag)
    target_git_repo.git.reset("--hard", "HEAD~1")

    # TODO: when available, switch this to use hvcs=none or similar config to avoid token use for push
    update_pyproject_toml(
        "tool.semantic_release.remote.ignore_token_for_push",
        True,
        target_repo_dir / pyproject_toml_file,
    )
    target_git_repo.git.commit(amend=True, no_edit=True, all=True)

    # push the current state to establish the remote (cannot push tags and branches at the same time)
    target_git_repo.git.push(remote_name, all=True)  # all branches
    target_git_repo.git.push(remote_name, tags=True)  # all tags

    # ensure bare remote HEAD points to the active branch so clones can checkout
    local_origin.git.symbolic_ref(
        "HEAD", f"refs/heads/{target_git_repo.active_branch.name}"
    )

    # Simulate CI environment after someone pushes to the repo
    ci_commit_sha = target_git_repo.head.commit.hexsha
    ci_branch = target_git_repo.active_branch.name

    # current remote tags
    remote_origin_tags_before = {tag.name for tag in local_origin.tags}

    # Simulate a CI environment by fetching the repo to a new location
    test_repo = Repo.init(str(example_project_dir / "ci_repo"))
    with test_repo.config_writer("repository") as config:
        config.set_value("core", "hookspath", "")
        config.set_value("commit", "gpgsign", False)
        config.set_value("tag", "gpgsign", False)

    # Configure and retrieve the repository (see GitHub actions/checkout@v5)
    test_repo.git.remote(
        "add",
        remote_name,
        f"file:///{PureWindowsPath(local_origin.working_dir).as_posix()}",
    )
    test_repo.git.fetch("--depth=1", remote_name, ci_commit_sha)

    # Simulate CI environment and recommended workflow (in docs)
    # NOTE: this could be done in 1 step, but most CI pipelines are doing it in 2 steps
    # 1. Checkout the commit sha (detached head)
    test_repo.git.checkout(ci_commit_sha, force=True)
    # 2. Forcefully set the branch to the current detached head
    test_repo.git.checkout("-B", ci_branch)

    # Act: run PSR on the cloned repo - it should verify upstream and succeed
    with temporary_working_directory(str(test_repo.working_dir)):
        cli_cmd = [MAIN_PROG_NAME, "--strict", VERSION_SUBCMD]
        result = run_cli(cli_cmd[1:], env={Github.DEFAULT_ENV_TOKEN_NAME: "1234"})

    remote_origin_tags_after = {tag.name for tag in local_origin.tags}

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)

    # Verify release occurred as expected
    with test_repo:
        assert latest_tag in test_repo.tags, "Expected release tag to be created"
        assert ci_commit_sha in [
            parent.hexsha for parent in test_repo.head.commit.parents
        ], "Expected new commit to be created on HEAD"
        different_tags = remote_origin_tags_after.difference(remote_origin_tags_before)
        assert latest_tag in different_tags, "Expected new tag to be pushed to remote"

    # Verify VCS release was created
    expected_vcs_url_post = 1
    assert expected_vcs_url_post == post_mocker.call_count  # one vcs release created


@pytest.mark.parametrize(
    "repo_fixture_name, build_repo_fn",
    [
        (
            repo_fixture_name,
            lazy_fixture(build_repo_fn_name),
        )
        for repo_fixture_name, build_repo_fn_name in [
            (
                repo_w_trunk_only_conventional_commits.__name__,
                build_trunk_only_repo_w_tags.__name__,
            ),
        ]
    ],
)
@pytest.mark.usefixtures(change_to_ex_proj_dir.__name__)
def test_version_no_upstream_check_on_no_version_commit(
    repo_fixture_name: str,
    run_cli: RunCliFn,
    build_repo_fn: BuildSpecificRepoFn,
    example_project_dir: ExProjectDir,
    git_repo_for_directory: GetGitRepo4DirFn,
    post_mocker: Mocker,
    get_cfg_value_from_def: GetCfgValueFromDefFn,
    get_versions_from_repo_build_def: GetVersionsFromRepoBuildDefFn,
    pyproject_toml_file: Path,
    update_pyproject_toml: UpdatePyprojectTomlFn,
):
    """
    Test that PSR succeeds when no version commit is needed, so the upstream check is skipped.

    This replicates the scenario that occurred on python-semantic-release/publish-action@v10.5.1
    where the version command was run and no version commit was needed, but it failed because
    it attempted to check the upstream branch anyway and we hard coded HEAD~1 because it expects
    a version commit to be created. This is the only reason why you would check the upstream branch
    because pushing a tag to the remote can happen even if the upstream branch has changed.
    """
    remote_name = "origin"
    # Create a bare remote (simulating origin)
    local_origin = Repo.init(str(example_project_dir / "local_origin"), bare=True)

    # build target repo into a temporary directory
    target_repo_dir = example_project_dir / repo_fixture_name
    commit_type: CommitConvention = (
        repo_fixture_name.split("commits", 1)[0].split("_")[-2]  # type: ignore[assignment]
    )
    target_repo_definition = build_repo_fn(
        repo_name=repo_fixture_name,
        commit_type=commit_type,
        dest_dir=target_repo_dir,
    )
    target_git_repo = git_repo_for_directory(target_repo_dir)

    # Configure the source repo to use the bare remote (removing any existing 'origin')
    with contextlib.suppress(AttributeError):
        target_git_repo.delete_remote(target_git_repo.remotes[remote_name])

    target_git_repo.create_remote(remote_name, str(local_origin.working_dir))

    # Remove last release before pushing to upstream
    tag_format_str = cast(
        "str", get_cfg_value_from_def(target_repo_definition, "tag_format_str")
    )
    latest_tag = tag_format_str.format(
        version=get_versions_from_repo_build_def(target_repo_definition)[-1]
    )
    target_git_repo.git.tag("-d", latest_tag)
    target_git_repo.git.reset("--hard", "HEAD~1")

    # Remove any version variables to ensure no version commit is needed
    update_pyproject_toml(
        "tool.semantic_release.version_variables",
        None,
        target_repo_dir / pyproject_toml_file,
    )
    update_pyproject_toml(
        "tool.semantic_release.version_toml",
        None,
        target_repo_dir / pyproject_toml_file,
    )
    # TODO: when available, switch this to use hvcs=none or similar config to avoid token use for push
    update_pyproject_toml(
        "tool.semantic_release.remote.ignore_token_for_push",
        True,
        target_repo_dir / pyproject_toml_file,
    )
    target_git_repo.git.commit(amend=True, no_edit=True, all=True)

    # push the current state to establish the remote (cannot push tags and branches at the same time)
    target_git_repo.git.push(remote_name, all=True)  # all branches
    target_git_repo.git.push(remote_name, tags=True)  # all tags

    # ensure bare remote HEAD points to the active branch so clones can checkout
    local_origin.git.symbolic_ref(
        "HEAD", f"refs/heads/{target_git_repo.active_branch.name}"
    )

    # Simulate CI environment after someone pushes to the repo
    ci_commit_sha = target_git_repo.head.commit.hexsha
    ci_branch = target_git_repo.active_branch.name

    # current remote tags
    remote_origin_tags_before = {tag.name for tag in local_origin.tags}

    # Simulate a CI environment by fetching the repo to a new location
    test_repo = Repo.init(str(example_project_dir / "ci_repo"))
    with test_repo.config_writer("repository") as config:
        config.set_value("core", "hookspath", "")
        config.set_value("commit", "gpgsign", False)
        config.set_value("tag", "gpgsign", False)

    # Configure and retrieve the repository (see GitHub actions/checkout@v5)
    test_repo.git.remote(
        "add",
        remote_name,
        f"file:///{PureWindowsPath(local_origin.working_dir).as_posix()}",
    )
    test_repo.git.fetch("--depth=1", remote_name, ci_commit_sha)

    # Simulate CI environment and recommended workflow (in docs)
    # NOTE: this could be done in 1 step, but most CI pipelines are doing it in 2 steps
    # 1. Checkout the commit sha (detached head)
    test_repo.git.checkout(ci_commit_sha, force=True)
    # 2. Forcefully set the branch to the current detached head
    test_repo.git.checkout("-B", ci_branch)

    # Act: run PSR on the cloned repo - it should verify upstream and succeed
    with temporary_working_directory(str(test_repo.working_dir)):
        # We don't use `--no-commit` here because we want to test that the upstream check is skipped
        # when PSR determines that no version commit is needed. If we used `--no-commit`, it would skip the
        # upstream check because it would think that a version commit was not needed.
        cli_cmd = [
            MAIN_PROG_NAME,
            "--strict",
            VERSION_SUBCMD,
            "--no-changelog",
            "--skip-build",
        ]
        result = run_cli(cli_cmd[1:], env={Github.DEFAULT_ENV_TOKEN_NAME: "1234"})

    remote_origin_tags_after = {tag.name for tag in local_origin.tags}

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)

    # Verify release occurred as expected
    with test_repo:
        assert latest_tag in test_repo.tags, "Expected release tag to be created"
        assert (
            ci_commit_sha == test_repo.head.commit.hexsha
        ), "Expected no new commit to be created on HEAD"
        different_tags = remote_origin_tags_after.difference(remote_origin_tags_before)
        assert latest_tag in different_tags, "Expected new tag to be pushed to remote"

    # Verify VCS release was created
    expected_vcs_url_post = 1
    assert expected_vcs_url_post == post_mocker.call_count  # one vcs release created


@pytest.mark.parametrize(
    "repo_fixture_name, build_repo_fn",
    [
        (
            repo_fixture_name,
            lazy_fixture(build_repo_fn_name),
        )
        for repo_fixture_name, build_repo_fn_name in [
            (
                repo_w_trunk_only_conventional_commits.__name__,
                build_trunk_only_repo_w_tags.__name__,
            ),
        ]
    ],
)
@pytest.mark.usefixtures(change_to_ex_proj_dir.__name__)
def test_version_upstream_check_fails_when_upstream_changed(
    repo_fixture_name: str,
    run_cli: RunCliFn,
    build_repo_fn: BuildSpecificRepoFn,
    example_project_dir: ExProjectDir,
    git_repo_for_directory: GetGitRepo4DirFn,
    post_mocker: Mocker,
    get_cfg_value_from_def: GetCfgValueFromDefFn,
    get_versions_from_repo_build_def: GetVersionsFromRepoBuildDefFn,
    pyproject_toml_file: Path,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    file_in_repo: str,
):
    remote_name = "origin"
    # Create a bare remote (simulating origin)
    local_origin = Repo.init(str(example_project_dir / "local_origin"), bare=True)

    # build target repo into a temporary directory
    target_repo_dir = example_project_dir / repo_fixture_name
    commit_type: CommitConvention = (
        repo_fixture_name.split("commits", 1)[0].split("_")[-2]  # type: ignore[assignment]
    )
    target_repo_definition = build_repo_fn(
        repo_name=repo_fixture_name,
        commit_type=commit_type,
        dest_dir=target_repo_dir,
    )
    target_git_repo = git_repo_for_directory(target_repo_dir)

    # Configure the source repo to use the bare remote (removing any existing 'origin')
    with contextlib.suppress(AttributeError):
        target_git_repo.delete_remote(target_git_repo.remotes[remote_name])

    target_git_repo.create_remote(remote_name, f"file://{local_origin.working_dir}")

    # Remove last release before pushing to upstream
    tag_format_str = cast(
        "str", get_cfg_value_from_def(target_repo_definition, "tag_format_str")
    )
    latest_tag = tag_format_str.format(
        version=get_versions_from_repo_build_def(target_repo_definition)[-1]
    )
    target_git_repo.git.tag("-d", latest_tag)
    target_git_repo.git.reset("--hard", "HEAD~1")

    # TODO: when available, switch this to use hvcs=none or similar config to avoid token use for push
    update_pyproject_toml(
        "tool.semantic_release.remote.ignore_token_for_push",
        True,
        target_repo_dir / pyproject_toml_file,
    )
    target_git_repo.git.commit(amend=True, no_edit=True, all=True)

    # push the current state to establish the remote (cannot push tags and branches at the same time)
    target_git_repo.git.push(remote_name, all=True)  # all branches
    target_git_repo.git.push(remote_name, tags=True)  # all tags

    # ensure bare remote HEAD points to the branch name used in the pushed repo
    local_origin.git.symbolic_ref(
        "HEAD", f"refs/heads/{target_git_repo.active_branch.name}"
    )

    # current remote tags
    remote_origin_tags_before = {tag.name for tag in local_origin.tags}

    # Clone the repo to simulate a local workspace
    test_repo = Repo.clone_from(
        f"file://{local_origin.working_dir}",
        str(example_project_dir / "repo_clone"),
        no_local=True,
    )
    with test_repo.config_writer("repository") as config:
        config.set_value("core", "hookspath", "")
        config.set_value("commit", "gpgsign", False)
        config.set_value("tag", "gpgsign", False)

    # Apply new commit to the original repo to simulate another developer pushing to upstream
    add_text_to_file(target_git_repo, str(target_repo_dir / file_in_repo))
    target_git_repo.index.add([str(file_in_repo)])
    target_git_repo.index.commit("feat: upstream change by another developer")
    target_git_repo.git.push(remote_name, target_git_repo.active_branch.name)

    # Act: run PSR - it should detect upstream changed and fail
    with temporary_working_directory(str(test_repo.working_dir)):
        cli_cmd = [MAIN_PROG_NAME, "--strict", VERSION_SUBCMD]
        result = run_cli(cli_cmd[1:], env={Github.DEFAULT_ENV_TOKEN_NAME: "1234"})

    remote_origin_tags_after = {tag.name for tag in local_origin.tags}

    # Evaluate
    assert_exit_code(1, result, cli_cmd)
    expected_err_msg = (
        f"Upstream branch '{remote_name}/{test_repo.active_branch.name}' has changed!"
    )
    # Verify error message mentions upstream
    assert (
        expected_err_msg in result.stderr
    ), f"Expected '{expected_err_msg}' in error output, got: {result.stderr}"

    assert (
        remote_origin_tags_before == remote_origin_tags_after
    ), "Expected no new tags to be pushed to remote"

    # Verify no VCS release was created
    expected_vcs_url_post = 0
    assert expected_vcs_url_post == post_mocker.call_count  # no vcs release created



from __future__ import annotations

from typing import TYPE_CHECKING

from semantic_release.commit_parser import ParsedCommit
from semantic_release.version.version import LevelBump

if TYPE_CHECKING:
    from tests.conftest import MakeCommitObjFn


def test_parsed_commit_computed_properties(make_commit_obj: MakeCommitObjFn):
    message = "feat(parser): Add new parser pattern"
    commit = make_commit_obj(message)

    parsed_commit = ParsedCommit(
        bump=LevelBump.MINOR,
        type="feature",
        scope="parser",
        descriptions=["Add new parser pattern"],
        breaking_descriptions=[],
        commit=commit,
    )

    assert message == parsed_commit.message
    assert commit.hexsha == parsed_commit.hexsha
    assert commit.hexsha[:7] == parsed_commit.short_hash



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, NamedTuple

import pytest
from git import Actor
from pytest_lazy_fixtures.lazy_fixture import lf as lazy_fixture

from semantic_release.changelog.release_history import ReleaseHistory
from semantic_release.version.translator import VersionTranslator
from semantic_release.version.version import Version

from tests.const import COMMIT_MESSAGE, CONVENTIONAL_COMMITS_MINOR
from tests.fixtures import (
    repo_w_git_flow_w_alpha_prereleases_n_conventional_commits,
    repo_w_git_flow_w_rc_n_alpha_prereleases_n_conventional_commits,
    repo_w_github_flow_w_feature_release_channel_conventional_commits,
    repo_w_no_tags_conventional_commits,
    repo_w_trunk_only_conventional_commits,
    repo_w_trunk_only_n_prereleases_conventional_commits,
)
from tests.util import add_text_to_file

if TYPE_CHECKING:
    from typing import Protocol

    from semantic_release.commit_parser.conventional import ConventionalCommitParser

    from tests.fixtures.git_repo import (
        BuiltRepoResult,
        GetCommitsFromRepoBuildDefFn,
        RepoDefinition,
    )

    class CreateReleaseHistoryFromRepoDefFn(Protocol):
        def __call__(self, repo_def: RepoDefinition) -> FakeReleaseHistoryElements: ...

# NOTE: not testing parser correctness here, just that the right commits end up
# in the right places. So we only compare that the commits with the messages
# we anticipate are in the right place, rather than by hash
# So we are only using the conventional parser


# We are also currently only testing that the "elements" key of the releases
# is correct, i.e. the commits are in the right place - the other fields
# will need special attention of their own later
class FakeReleaseHistoryElements(NamedTuple):
    """
    A fake release history structure that abstracts away the Parser-specific
    logic and only focuses that the commit messages are in the correct order and place.

    Where generally a ParsedCommit object exists, here we just use the actual `commit.message`.
    """

    unreleased: dict[str, list[str]]
    released: dict[Version, dict[str, list[str]]]


@pytest.fixture(scope="session")
def create_release_history_from_repo_def() -> CreateReleaseHistoryFromRepoDefFn:
    def _create_release_history_from_repo_def(
        repo_def: RepoDefinition,
    ) -> FakeReleaseHistoryElements:
        # Organize the commits into the expected structure
        unreleased_history = {}
        released_history = {}
        for version_str, version_def in repo_def.items():
            commits_per_group: dict[str, list] = {
                "Unknown": [],
            }

            for commit in version_def["commits"]:
                if commit["category"] not in commits_per_group:
                    commits_per_group[commit["category"]] = []

                commits_per_group[commit["category"]].append(commit["msg"].strip())

            if version_str == "Unreleased":
                unreleased_history = commits_per_group
                continue

            # handle released versions
            version = Version.parse(version_str)

            # add the PSR version commit message
            commits_per_group["Unknown"].append(
                COMMIT_MESSAGE.format(version=version).strip()
            )

            # store the organized commits for this version
            released_history[version] = commits_per_group

        return FakeReleaseHistoryElements(
            unreleased=unreleased_history,
            released=released_history,
        )

    return _create_release_history_from_repo_def


@pytest.mark.parametrize(
    "repo_result",
    [
        # CONVENTIONAL parser
        lazy_fixture(repo_w_no_tags_conventional_commits.__name__),
        *[
            pytest.param(
                lazy_fixture(repo_fixture_name),
                marks=pytest.mark.comprehensive,
            )
            for repo_fixture_name in [
                repo_w_trunk_only_conventional_commits.__name__,
                repo_w_trunk_only_n_prereleases_conventional_commits.__name__,
                # This is not tested because currently unable to disern the commits that were squashed or not
                # repo_w_github_flow_w_default_release_channel_conventional_commits.__name__,
                repo_w_github_flow_w_feature_release_channel_conventional_commits.__name__,
                repo_w_git_flow_w_alpha_prereleases_n_conventional_commits.__name__,
                repo_w_git_flow_w_rc_n_alpha_prereleases_n_conventional_commits.__name__,
            ]
        ],
    ],
)
@pytest.mark.order("last")
def test_release_history(
    repo_result: BuiltRepoResult,
    default_conventional_parser: ConventionalCommitParser,
    file_in_repo: str,
    create_release_history_from_repo_def: CreateReleaseHistoryFromRepoDefFn,
    get_commits_from_repo_build_def: GetCommitsFromRepoBuildDefFn,
):
    repo = repo_result["repo"]
    expected_release_history = create_release_history_from_repo_def(
        get_commits_from_repo_build_def(
            repo_result["definition"],
            ignore_merge_commits=default_conventional_parser.options.ignore_merge_commits,
        )
    )
    expected_released_versions = sorted(
        map(str, expected_release_history.released.keys())
    )

    translator = VersionTranslator()
    # Nothing has unreleased commits currently
    history = ReleaseHistory.from_git_history(
        repo,
        translator,
        default_conventional_parser,  # type: ignore[arg-type]
    )
    released = history.released

    actual_released_versions = sorted(map(str, released.keys()))
    assert expected_released_versions == actual_released_versions

    for k in expected_release_history.released:
        expected = expected_release_history.released[k]
        expected_released_messages = str.join(
            "\n---\n", sorted([msg for bucket in expected.values() for msg in bucket])
        )

        actual = released[k]["elements"]
        actual_released_messages = str.join(
            "\n---\n",
            sorted(
                [
                    str(res.commit.message)
                    for results in actual.values()
                    for res in results
                ]
            ),
        )
        assert expected_released_messages == actual_released_messages

    # PART 2: add some commits to the repo and check that they are in the right place

    for commit_message in CONVENTIONAL_COMMITS_MINOR:
        add_text_to_file(repo, file_in_repo)
        repo.git.commit(m=commit_message)

    expected_unreleased_messages = str.join(
        "\n---\n",
        sorted(
            [
                str(msg).strip()
                for bucket in [
                    CONVENTIONAL_COMMITS_MINOR[::-1],
                    *expected_release_history.unreleased.values(),
                ]
                for msg in bucket
            ]
        ),
    )

    # Now we should have some unreleased commits, and nothing new released
    new_history = ReleaseHistory.from_git_history(
        repo,
        translator,
        default_conventional_parser,  # type: ignore[arg-type]
    )
    new_unreleased = new_history.unreleased
    new_released = new_history.released

    actual_unreleased_messages = str.join(
        "\n---\n",
        sorted(
            [
                str(res.commit.message)
                for results in new_unreleased.values()
                for res in results
            ]
        ),
    )

    assert expected_unreleased_messages == actual_unreleased_messages
    assert (
        new_released == released
    ), "something that shouldn't be considered release has been released"


@pytest.mark.parametrize(
    "repo_result",
    [
        lazy_fixture(repo_w_no_tags_conventional_commits.__name__),
        *[
            pytest.param(
                lazy_fixture(repo_fixture_name),
                marks=pytest.mark.comprehensive,
            )
            for repo_fixture_name in [
                repo_w_trunk_only_conventional_commits.__name__,
                repo_w_trunk_only_n_prereleases_conventional_commits.__name__,
                repo_w_github_flow_w_feature_release_channel_conventional_commits.__name__,
                repo_w_git_flow_w_alpha_prereleases_n_conventional_commits.__name__,
                repo_w_git_flow_w_rc_n_alpha_prereleases_n_conventional_commits.__name__,
            ]
        ],
    ],
)
@pytest.mark.order("last")
def test_release_history_releases(
    repo_result: BuiltRepoResult, default_conventional_parser: ConventionalCommitParser
):
    new_version = Version.parse("100.10.1")
    actor = Actor("semantic-release", "semantic-release")
    release_history = ReleaseHistory.from_git_history(
        repo=repo_result["repo"],
        translator=VersionTranslator(),
        commit_parser=default_conventional_parser,  # type: ignore[arg-type]
    )
    tagged_date = datetime.now()
    new_rh = release_history.release(
        new_version,
        committer=actor,
        tagger=actor,
        tagged_date=tagged_date,
    )

    assert new_rh is not release_history
    assert new_rh.unreleased == {}
    assert new_rh.released == {
        new_version: {
            "tagger": actor,
            "committer": actor,
            "tagged_date": tagged_date,
            "elements": release_history.unreleased,
            "version": new_version,
        },
        **release_history.released,
    }


@pytest.mark.parametrize(
    "repo_result",
    [
        lazy_fixture(repo_w_no_tags_conventional_commits.__name__),
        *[
            pytest.param(
                lazy_fixture(repo_fixture_name),
                marks=pytest.mark.comprehensive,
            )
            for repo_fixture_name in [
                repo_w_trunk_only_conventional_commits.__name__,
                repo_w_trunk_only_n_prereleases_conventional_commits.__name__,
                repo_w_github_flow_w_feature_release_channel_conventional_commits.__name__,
                repo_w_git_flow_w_alpha_prereleases_n_conventional_commits.__name__,
                repo_w_git_flow_w_rc_n_alpha_prereleases_n_conventional_commits.__name__,
            ]
        ],
    ],
)
@pytest.mark.order("last")
def test_all_matching_repo_tags_are_released(
    repo_result: BuiltRepoResult, default_conventional_parser: ConventionalCommitParser
):
    repo = repo_result["repo"]
    translator = VersionTranslator()
    release_history = ReleaseHistory.from_git_history(
        repo=repo,
        translator=translator,
        commit_parser=default_conventional_parser,  # type: ignore[arg-type]
    )

    for tag in repo.tags:
        assert translator.from_tag(tag.name) in release_history.released



from __future__ import annotations

import os
from pathlib import Path
from re import IGNORECASE, MULTILINE, compile as regexp
from typing import TYPE_CHECKING
from unittest.mock import MagicMock

import git.remote as git_remote
import pytest
from requests_mock import ANY

from semantic_release.cli import config as cli_config_module
from semantic_release.cli.config import (
    GlobalCommandLineOptions,
    RawConfig,
    RuntimeContext,
)
from semantic_release.cli.const import DEFAULT_CONFIG_FILE
from semantic_release.cli.util import load_raw_config_file

from tests.util import prepare_mocked_git_command_wrapper_type

if TYPE_CHECKING:
    from re import Pattern
    from typing import Protocol

    from git.repo import Repo
    from pytest import MonkeyPatch
    from requests_mock.mocker import Mocker

    from tests.fixtures.example_project import ExProjectDir

    class GetSanitizedChangelogContentFn(Protocol):
        def __call__(
            self,
            repo_dir: Path,
            changelog_file: Path = ...,
            remove_insertion_flag: bool = True,
        ) -> str: ...

    class ReadConfigFileFn(Protocol):
        """Read the raw config file from `config_path`."""

        def __call__(self, file: Path | str) -> RawConfig: ...

    class RetrieveRuntimeContextFn(Protocol):
        """Retrieve the runtime context for a repo."""

        def __call__(self, repo: Repo) -> RuntimeContext: ...

    class StripLoggingMessagesFn(Protocol):
        def __call__(self, log: str) -> str: ...


@pytest.hookimpl(tryfirst=True)
def pytest_collection_modifyitems(items: list[pytest.Item]) -> None:
    """Apply the e2e marker to all tests in the end-to-end test directory."""
    cli_test_directory = Path(__file__).parent
    for item in items:
        if cli_test_directory in item.path.parents:
            item.add_marker(pytest.mark.e2e)


@pytest.fixture
def post_mocker(requests_mock: Mocker) -> Mocker:
    """Patch all POST requests, mocking a response body for VCS release creation."""
    requests_mock.register_uri("POST", ANY, json={"id": 999})
    return requests_mock


@pytest.fixture
def mocked_git_push(monkeypatch: MonkeyPatch) -> MagicMock:
    """Mock the `Repo.git.push()` method in `semantic_release.cli.main`."""
    mocked_push = MagicMock()
    cls = prepare_mocked_git_command_wrapper_type(push=mocked_push)
    monkeypatch.setattr(cli_config_module.Repo, "GitCommandWrapperType", cls)
    return mocked_push


@pytest.fixture
def mocked_git_fetch(monkeypatch: MonkeyPatch) -> MagicMock:
    """
    Mock the `Repo.git.fetch()` method in `semantic_release.cli.main` and
    `git.Repo.remotes.Remote.fetch()`.
    """
    mocked_fetch = MagicMock()
    cls = prepare_mocked_git_command_wrapper_type(fetch=mocked_fetch)
    monkeypatch.setattr(cli_config_module.Repo, "GitCommandWrapperType", cls)

    # define a small wrapper so the MagicMock does not receive `self`
    def _fetch(self, *args, **kwargs):
        return mocked_fetch(*args, **kwargs)

    # Replace the method on the Remote class used by GitPython
    monkeypatch.setattr(git_remote.Remote, "fetch", _fetch, raising=True)

    return mocked_fetch


@pytest.fixture
def config_path(example_project_dir: ExProjectDir) -> Path:
    return example_project_dir / DEFAULT_CONFIG_FILE


@pytest.fixture(scope="session")
def read_config_file() -> ReadConfigFileFn:
    def _read_config_file(file: Path | str) -> RawConfig:
        config_text = load_raw_config_file(file)
        return RawConfig.model_validate(config_text)

    return _read_config_file


@pytest.fixture
def cli_options(config_path: Path) -> GlobalCommandLineOptions:
    return GlobalCommandLineOptions(
        noop=False,
        verbosity=0,
        strict=False,
        config_file=str(config_path),
    )


@pytest.fixture
def retrieve_runtime_context(
    read_config_file: ReadConfigFileFn,
    cli_options: GlobalCommandLineOptions,
) -> RetrieveRuntimeContextFn:
    def _retrieve_runtime_context(repo: Repo) -> RuntimeContext:
        cwd = os.getcwd()
        repo_dir = str(Path(repo.working_dir).resolve())

        os.chdir(repo_dir)
        try:
            raw_config = read_config_file(cli_options.config_file)
            return RuntimeContext.from_raw_config(raw_config, cli_options)
        finally:
            os.chdir(cwd)

    return _retrieve_runtime_context


@pytest.fixture(scope="session")
def strip_logging_messages() -> StripLoggingMessagesFn:
    """Fixture to strip logging messages from the output."""
    # Log levels match SemanticReleaseLogLevel enum values
    logger_msg_pattern = regexp(
        r"^\s*(?:\[\d\d:\d\d:\d\d\])?\s*(FATAL|CRITICAL|ERROR|WARNING|INFO|DEBUG|SILLY).*?\n(?:\s+\S.*?\n)*(?!\n[ ]{11})",
        MULTILINE,
    )

    def _strip_logging_messages(log: str) -> str:
        # Make sure it ends with a newline
        return logger_msg_pattern.sub("", log.rstrip("\n") + "\n")

    return _strip_logging_messages


@pytest.fixture(scope="session")
def long_hash_pattern() -> Pattern[str]:
    return regexp(r"\b([0-9a-f]{40})\b", IGNORECASE)


@pytest.fixture(scope="session")
def short_hash_pattern() -> Pattern[str]:
    return regexp(r"\b([0-9a-f]{7})\b", IGNORECASE)


@pytest.fixture(scope="session")
def get_sanitized_rst_changelog_content(
    changelog_rst_file: Path,
    default_rst_changelog_insertion_flag: str,
    long_hash_pattern: Pattern[str],
    short_hash_pattern: Pattern[str],
) -> GetSanitizedChangelogContentFn:
    rst_short_hash_link_pattern = regexp(r"(_[0-9a-f]{7})\b", IGNORECASE)

    def _get_sanitized_rst_changelog_content(
        repo_dir: Path,
        changelog_file: Path = changelog_rst_file,
        remove_insertion_flag: bool = False,
    ) -> str:
        if not (changelog_path := repo_dir / changelog_file).exists():
            return ""

        # Note that our repo generation fixture includes the insertion flag automatically
        # toggle remove_insertion_flag to True to remove the insertion flag, applies to Init mode repos
        with changelog_path.open(newline=os.linesep) as rfd:
            # use os.linesep here because the insertion flag is os-specific
            # but convert the content to universal newlines for comparison
            changelog_content = (
                rfd.read().replace(
                    f"{default_rst_changelog_insertion_flag}{os.linesep}", ""
                )
                if remove_insertion_flag
                else rfd.read()
            ).replace("\r", "")

        changelog_content = long_hash_pattern.sub("0" * 40, changelog_content)
        changelog_content = short_hash_pattern.sub("0" * 7, changelog_content)
        return rst_short_hash_link_pattern.sub(f'_{"0" * 7}', changelog_content)

    return _get_sanitized_rst_changelog_content


@pytest.fixture(scope="session")
def get_sanitized_md_changelog_content(
    changelog_md_file: Path,
    default_md_changelog_insertion_flag: str,
    long_hash_pattern: Pattern[str],
    short_hash_pattern: Pattern[str],
) -> GetSanitizedChangelogContentFn:
    def _get_sanitized_md_changelog_content(
        repo_dir: Path,
        changelog_file: Path = changelog_md_file,
        remove_insertion_flag: bool = False,
    ) -> str:
        if not (changelog_path := repo_dir / changelog_file).exists():
            return ""

        # Note that our repo generation fixture includes the insertion flag automatically
        # toggle remove_insertion_flag to True to remove the insertion flag, applies to Init mode repos
        with changelog_path.open(newline=os.linesep) as rfd:
            # use os.linesep here because the insertion flag is os-specific
            # but convert the content to universal newlines for comparison
            changelog_content = (
                rfd.read().replace(
                    f"{default_md_changelog_insertion_flag}{os.linesep}", ""
                )
                if remove_insertion_flag
                else rfd.read()
            ).replace("\r", "")

        changelog_content = long_hash_pattern.sub("0" * 40, changelog_content)
        return short_hash_pattern.sub("0" * 7, changelog_content)

    return _get_sanitized_md_changelog_content



from __future__ import annotations

import os
from datetime import datetime, timezone
from typing import TYPE_CHECKING

import pytest
from freezegun import freeze_time
from pytest_lazy_fixtures.lazy_fixture import lf as lazy_fixture

from semantic_release.changelog.context import ChangelogMode
from semantic_release.cli.config import ChangelogOutputFormat

from tests.const import MAIN_PROG_NAME, VERSION_SUBCMD
from tests.fixtures.example_project import (
    default_md_changelog_insertion_flag,
    default_rst_changelog_insertion_flag,
    example_changelog_md,
    example_changelog_rst,
)
from tests.fixtures.repos import (
    repo_w_git_flow_w_alpha_prereleases_n_conventional_commits,
    repo_w_git_flow_w_alpha_prereleases_n_emoji_commits,
    repo_w_git_flow_w_alpha_prereleases_n_scipy_commits,
    repo_w_git_flow_w_rc_n_alpha_prereleases_n_conventional_commits,
    repo_w_git_flow_w_rc_n_alpha_prereleases_n_conventional_commits_using_tag_format,
    repo_w_git_flow_w_rc_n_alpha_prereleases_n_emoji_commits,
    repo_w_git_flow_w_rc_n_alpha_prereleases_n_scipy_commits,
    repo_w_github_flow_w_default_release_channel_conventional_commits,
    repo_w_github_flow_w_default_release_channel_emoji_commits,
    repo_w_github_flow_w_default_release_channel_scipy_commits,
    repo_w_github_flow_w_feature_release_channel_conventional_commits,
    repo_w_github_flow_w_feature_release_channel_emoji_commits,
    repo_w_github_flow_w_feature_release_channel_scipy_commits,
    repo_w_no_tags_conventional_commits,
    repo_w_no_tags_conventional_commits_unmasked_initial_release,
    repo_w_no_tags_emoji_commits,
    repo_w_no_tags_scipy_commits,
    repo_w_trunk_only_conventional_commits,
    repo_w_trunk_only_emoji_commits,
    repo_w_trunk_only_n_prereleases_conventional_commits,
    repo_w_trunk_only_n_prereleases_emoji_commits,
    repo_w_trunk_only_n_prereleases_scipy_commits,
    repo_w_trunk_only_scipy_commits,
)
from tests.util import assert_successful_exit_code

if TYPE_CHECKING:
    from pathlib import Path

    from tests.conftest import (
        FormatDateStrFn,
        GetCachedRepoDataFn,
        GetStableDateNowFn,
        RunCliFn,
    )
    from tests.fixtures.example_project import UpdatePyprojectTomlFn
    from tests.fixtures.git_repo import (
        BuiltRepoResult,
        CommitConvention,
        GetCommitsFromRepoBuildDefFn,
        GetVersionsFromRepoBuildDefFn,
    )


@pytest.mark.parametrize(
    "changelog_file, insertion_flag",
    [
        (
            # ChangelogOutputFormat.MARKDOWN
            lazy_fixture(example_changelog_md.__name__),
            lazy_fixture(default_md_changelog_insertion_flag.__name__),
        ),
        (
            # ChangelogOutputFormat.RESTRUCTURED_TEXT
            lazy_fixture(example_changelog_rst.__name__),
            lazy_fixture(default_rst_changelog_insertion_flag.__name__),
        ),
    ],
)
@pytest.mark.parametrize(
    "repo_result, repo_fixture_name, tag_format",
    [
        (
            lazy_fixture(repo_w_trunk_only_conventional_commits.__name__),
            repo_w_trunk_only_conventional_commits.__name__,
            "v{version}",
        ),
        *[
            pytest.param(
                lazy_fixture(repo_fixture),
                repo_fixture,
                "v{version}" if tag_format is None else tag_format,
                marks=pytest.mark.comprehensive,
            )
            for repo_fixture, tag_format in [
                # Must have a previous release/tag
                *[
                    (
                        repo_fixture_name,
                        None,
                    )
                    for repo_fixture_name in [
                        # repo_with_single_branch_conventional_commits.__name__, # default
                        repo_w_trunk_only_emoji_commits.__name__,
                        repo_w_trunk_only_scipy_commits.__name__,
                    ]
                ],
                *[
                    (
                        repo_fixture_name,
                        None,
                    )
                    for repo_fixture_name in [
                        repo_w_trunk_only_n_prereleases_conventional_commits.__name__,
                        repo_w_trunk_only_n_prereleases_emoji_commits.__name__,
                        repo_w_trunk_only_n_prereleases_scipy_commits.__name__,
                    ]
                ],
                *[
                    (
                        repo_fixture_name,
                        None,
                    )
                    for repo_fixture_name in [
                        repo_w_github_flow_w_default_release_channel_conventional_commits.__name__,
                        repo_w_github_flow_w_default_release_channel_emoji_commits.__name__,
                        repo_w_github_flow_w_default_release_channel_scipy_commits.__name__,
                    ]
                ],
                *[
                    (
                        repo_fixture_name,
                        None,
                    )
                    for repo_fixture_name in [
                        repo_w_github_flow_w_feature_release_channel_conventional_commits.__name__,
                        repo_w_github_flow_w_feature_release_channel_emoji_commits.__name__,
                        repo_w_github_flow_w_feature_release_channel_scipy_commits.__name__,
                    ]
                ],
                *[
                    (
                        repo_fixture_name,
                        None,
                    )
                    for repo_fixture_name in [
                        repo_w_git_flow_w_alpha_prereleases_n_conventional_commits.__name__,
                        repo_w_git_flow_w_alpha_prereleases_n_emoji_commits.__name__,
                        repo_w_git_flow_w_alpha_prereleases_n_scipy_commits.__name__,
                    ]
                ],
                *[
                    (
                        repo_fixture_name,
                        None,
                    )
                    for repo_fixture_name in [
                        repo_w_git_flow_w_rc_n_alpha_prereleases_n_conventional_commits.__name__,
                        repo_w_git_flow_w_rc_n_alpha_prereleases_n_emoji_commits.__name__,
                        repo_w_git_flow_w_rc_n_alpha_prereleases_n_scipy_commits.__name__,
                    ]
                ],
                *[
                    (
                        repo_fixture_name,
                        "submod-v{version}",
                    )
                    for repo_fixture_name in [
                        repo_w_git_flow_w_rc_n_alpha_prereleases_n_conventional_commits_using_tag_format.__name__,
                    ]
                ],
            ]
        ],
    ],
)
def test_version_updates_changelog_w_new_version(
    repo_result: BuiltRepoResult,
    get_versions_from_repo_build_def: GetVersionsFromRepoBuildDefFn,
    tag_format: str,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    run_cli: RunCliFn,
    changelog_file: Path,
    insertion_flag: str,
    repo_fixture_name: str,
    stable_now_date: GetStableDateNowFn,
    get_cached_repo_data: GetCachedRepoDataFn,
):
    """
    Given a previously released custom modified changelog file,
    When the version command is run with changelog.mode set to "update",
    Then the version is created and the changelog file is updated with new release info
        while maintaining the previously customized content
    """
    repo = repo_result["repo"]
    latest_tag = tag_format.format(
        version=get_versions_from_repo_build_def(repo_result["definition"])[-1]
    )

    if not (repo_build_data := get_cached_repo_data(repo_fixture_name)):
        pytest.fail("Repo build date not found in cache")

    repo_build_datetime = datetime.strptime(repo_build_data["build_date"], "%Y-%m-%d")
    now_datetime = stable_now_date().replace(
        year=repo_build_datetime.year,
        month=repo_build_datetime.month,
        day=repo_build_datetime.day,
    )

    # Custom text to maintain (must be different from the default)
    custom_text = "---{ls}{ls}Custom footer text{ls}".format(ls=os.linesep)

    # Capture expected changelog content
    with changelog_file.open(newline=os.linesep) as rfd:
        initial_changelog_parts = rfd.read().split(insertion_flag)

    expected_changelog_content = str.join(
        insertion_flag,
        [
            initial_changelog_parts[0],
            str.join(
                os.linesep,
                [
                    initial_changelog_parts[1],
                    "",
                    custom_text,
                ],
            ),
        ],
    )

    # Reverse last release
    repo.git.tag("-d", latest_tag)
    repo.git.reset("--hard", "HEAD~1")

    # Set the project configurations
    update_pyproject_toml(
        "tool.semantic_release.changelog.mode", ChangelogMode.UPDATE.value
    )
    update_pyproject_toml(
        "tool.semantic_release.changelog.default_templates.changelog_file",
        str(changelog_file.name),
    )

    # Modify the current changelog with our custom text at bottom
    # Universal newlines is ok here since we are writing it back out
    # and not working with the os-specific insertion flag
    changelog_file.write_text(
        str.join(
            "\n",
            [
                changelog_file.read_text(),
                "",
                custom_text,
            ],
        )
    )

    with freeze_time(now_datetime.astimezone(timezone.utc)):
        cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, "--no-push", "--changelog"]
        result = run_cli(cli_cmd[1:])

    # Capture the new changelog content (os aware because of expected content)
    with changelog_file.open(newline=os.linesep) as rfd:
        actual_content = rfd.read()

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)
    assert expected_changelog_content == actual_content


@pytest.mark.parametrize(
    "changelog_format, changelog_file, insertion_flag",
    [
        (
            ChangelogOutputFormat.MARKDOWN,
            lazy_fixture(example_changelog_md.__name__),
            lazy_fixture(default_md_changelog_insertion_flag.__name__),
        ),
        (
            ChangelogOutputFormat.RESTRUCTURED_TEXT,
            lazy_fixture(example_changelog_rst.__name__),
            lazy_fixture(default_rst_changelog_insertion_flag.__name__),
        ),
    ],
)
@pytest.mark.parametrize(
    "repo_result, repo_fixture_name",
    [
        (
            lazy_fixture(repo_w_no_tags_conventional_commits.__name__),
            repo_w_no_tags_conventional_commits.__name__,
        ),
        *[
            pytest.param(
                lazy_fixture(repo_fixture),
                repo_fixture,
                marks=pytest.mark.comprehensive,
            )
            for repo_fixture in [
                # Must not have a single release/tag
                # repo_with_no_tags_conventional_commits.__name__, # default
                repo_w_no_tags_emoji_commits.__name__,
                repo_w_no_tags_scipy_commits.__name__,
            ]
        ],
    ],
)
def test_version_updates_changelog_wo_prev_releases(
    repo_result: BuiltRepoResult,
    repo_fixture_name: str,
    run_cli: RunCliFn,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    changelog_format: ChangelogOutputFormat,
    changelog_file: Path,
    insertion_flag: str,
    stable_now_date: GetStableDateNowFn,
    format_date_str: FormatDateStrFn,
    get_cached_repo_data: GetCachedRepoDataFn,
):
    """
    Given the repository has no releases and the user has provided a initialized changelog,
    When the version command is run with changelog.mode set to "update",
    Then the version is created and the changelog file is updated with only an initial release statement
    """
    if not (repo_build_data := get_cached_repo_data(repo_fixture_name)):
        pytest.fail("Repo build date not found in cache")

    repo_build_datetime = datetime.strptime(repo_build_data["build_date"], "%Y-%m-%d")
    now_datetime = stable_now_date().replace(
        year=repo_build_datetime.year,
        month=repo_build_datetime.month,
        day=repo_build_datetime.day,
    )
    repo_build_date_str = format_date_str(now_datetime)

    # Custom text to maintain (must be different from the default)
    custom_text = "---{ls}{ls}Custom footer text{ls}".format(ls=os.linesep)

    # Set the project configurations
    update_pyproject_toml(
        "tool.semantic_release.changelog.mode", ChangelogMode.UPDATE.value
    )
    update_pyproject_toml(
        "tool.semantic_release.changelog.default_templates.changelog_file",
        str(changelog_file.name),
    )

    version = "v1.0.0"
    rst_version_header = f"{version} ({repo_build_date_str})"
    txt_after_insertion_flag = {
        ChangelogOutputFormat.MARKDOWN: str.join(
            os.linesep,
            [
                f"## {version} ({repo_build_date_str})",
                "",
                "- Initial Release",
            ],
        ),
        ChangelogOutputFormat.RESTRUCTURED_TEXT: str.join(
            os.linesep,
            [
                f".. _changelog-{version}:",
                "",
                rst_version_header,
                f"{'=' * len(rst_version_header)}",
                "",
                "* Initial Release",
            ],
        ),
    }

    # Capture and modify the current changelog content to become the expected output
    # We much use os.linesep here since the insertion flag is os-specific
    with changelog_file.open(newline=os.linesep) as rfd:
        initial_changelog_parts = rfd.read().split(insertion_flag)

    # content is os-specific because of the insertion flag & how we read the original file
    expected_changelog_content = str.join(
        insertion_flag,
        [
            initial_changelog_parts[0],
            str.join(
                os.linesep,
                [
                    os.linesep,
                    txt_after_insertion_flag[changelog_format],
                    "",
                    custom_text,
                ],
            ),
        ],
    )

    # Grab the Unreleased changelog & create the initialized user changelog
    # force output to not perform any newline translations
    with changelog_file.open(mode="w", newline="") as wfd:
        wfd.write(
            str.join(
                insertion_flag,
                [initial_changelog_parts[0], f"{os.linesep * 2}{custom_text}"],
            )
        )
        wfd.flush()

    # Act
    with freeze_time(now_datetime.astimezone(timezone.utc)):
        cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, "--no-push", "--changelog"]
        result = run_cli(cli_cmd[1:])

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)

    # Ensure changelog exists
    assert changelog_file.exists()

    # Capture the new changelog content (os aware because of expected content)
    with changelog_file.open(newline=os.linesep) as rfd:
        actual_content = rfd.read()

    # Check that the changelog footer is maintained and updated with Unreleased info
    assert expected_changelog_content == actual_content


@pytest.mark.parametrize(
    "changelog_format, changelog_file, insertion_flag",
    [
        (
            ChangelogOutputFormat.MARKDOWN,
            lazy_fixture(example_changelog_md.__name__),
            lazy_fixture(default_md_changelog_insertion_flag.__name__),
        ),
        (
            ChangelogOutputFormat.RESTRUCTURED_TEXT,
            lazy_fixture(example_changelog_rst.__name__),
            lazy_fixture(default_rst_changelog_insertion_flag.__name__),
        ),
    ],
)
@pytest.mark.parametrize(
    "repo_result, repo_fixture_name",
    [
        pytest.param(
            lazy_fixture(repo_fixture),
            repo_fixture,
            marks=pytest.mark.comprehensive,
        )
        for repo_fixture in [
            # Must not have a single release/tag
            repo_w_no_tags_conventional_commits_unmasked_initial_release.__name__,
        ]
    ],
)
def test_version_updates_changelog_wo_prev_releases_n_unmasked_initial_release(
    repo_result: BuiltRepoResult,
    repo_fixture_name: str,
    run_cli: RunCliFn,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    changelog_format: ChangelogOutputFormat,
    changelog_file: Path,
    insertion_flag: str,
    stable_now_date: GetStableDateNowFn,
    format_date_str: FormatDateStrFn,
    get_cached_repo_data: GetCachedRepoDataFn,
):
    """
    Given the repository has no releases and the user has provided a initialized changelog,
    When the version command is run with changelog.mode set to "update",
    Then the version is created and the changelog file is updated with new release info
    """
    if not (repo_build_data := get_cached_repo_data(repo_fixture_name)):
        pytest.fail("Repo build date not found in cache")

    repo_build_datetime = datetime.strptime(repo_build_data["build_date"], "%Y-%m-%d")
    now_datetime = stable_now_date().replace(
        year=repo_build_datetime.year,
        month=repo_build_datetime.month,
        day=repo_build_datetime.day,
    )
    repo_build_date_str = format_date_str(now_datetime)

    # Custom text to maintain (must be different from the default)
    custom_text = "---{ls}{ls}Custom footer text{ls}".format(ls=os.linesep)

    # Set the project configurations
    update_pyproject_toml(
        "tool.semantic_release.changelog.mode", ChangelogMode.UPDATE.value
    )
    update_pyproject_toml(
        "tool.semantic_release.changelog.default_templates.changelog_file",
        str(changelog_file.name),
    )

    version = "v1.0.0"
    rst_version_header = f"{version} ({repo_build_date_str})"
    search_n_replacements = {
        ChangelogOutputFormat.MARKDOWN: (
            "## Unreleased",
            f"## {version} ({repo_build_date_str})",
        ),
        ChangelogOutputFormat.RESTRUCTURED_TEXT: (
            ".. _changelog-unreleased:{ls}{ls}Unreleased{ls}{underline}".format(
                ls=os.linesep,
                underline="=" * len("Unreleased"),
            ),
            str.join(
                os.linesep,
                [
                    f".. _changelog-{version}:",
                    "",
                    rst_version_header,
                    f"{'=' * len(rst_version_header)}",
                ],
            ),
        ),
    }

    search_text = search_n_replacements[changelog_format][0]
    replacement_text = search_n_replacements[changelog_format][1]

    # Capture and modify the current changelog content to become the expected output
    # We much use os.linesep here since the insertion flag is os-specific
    with changelog_file.open(newline=os.linesep) as rfd:
        initial_changelog_parts = rfd.read().split(insertion_flag)

    # content is os-specific because of the insertion flag & how we read the original file
    expected_changelog_content = str.join(
        insertion_flag,
        [
            initial_changelog_parts[0],
            str.join(
                os.linesep,
                [
                    initial_changelog_parts[1].replace(
                        search_text,
                        replacement_text,
                    ),
                    "",
                    custom_text,
                ],
            ),
        ],
    )

    # Grab the Unreleased changelog & create the initialized user changelog
    # force output to not perform any newline translations
    with changelog_file.open(mode="w", newline="") as wfd:
        wfd.write(
            str.join(
                insertion_flag,
                [initial_changelog_parts[0], f"{os.linesep * 2}{custom_text}"],
            )
        )
        wfd.flush()

    # Act
    with freeze_time(now_datetime.astimezone(timezone.utc)):
        cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, "--no-push", "--changelog"]
        result = run_cli(cli_cmd[1:])

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)

    # Ensure changelog exists
    assert changelog_file.exists()

    # Capture the new changelog content (os aware because of expected content)
    with changelog_file.open(newline=os.linesep) as rfd:
        actual_content = rfd.read()

    # Check that the changelog footer is maintained and updated with Unreleased info
    assert expected_changelog_content == actual_content


@pytest.mark.parametrize(
    "changelog_file",
    [
        lazy_fixture(example_changelog_md.__name__),
        lazy_fixture(example_changelog_rst.__name__),
    ],
)
@pytest.mark.parametrize(
    "repo_result, repo_fixture_name, tag_format",
    [
        (
            lazy_fixture(repo_w_trunk_only_conventional_commits.__name__),
            repo_w_trunk_only_conventional_commits.__name__,
            "v{version}",
        ),
        *[
            pytest.param(
                lazy_fixture(repo_fixture),
                repo_fixture,
                "v{version}" if tag_format is None else tag_format,
                marks=pytest.mark.comprehensive,
            )
            for repo_fixture, tag_format in [
                # Must have a previous release/tag
                *[
                    (
                        repo_fixture_name,
                        None,
                    )
                    for repo_fixture_name in [
                        # repo_with_single_branch_conventional_commits.__name__, # default
                        repo_w_trunk_only_emoji_commits.__name__,
                        repo_w_trunk_only_scipy_commits.__name__,
                    ]
                ],
                *[
                    (
                        repo_fixture_name,
                        None,
                    )
                    for repo_fixture_name in [
                        repo_w_trunk_only_n_prereleases_conventional_commits.__name__,
                        repo_w_trunk_only_n_prereleases_emoji_commits.__name__,
                        repo_w_trunk_only_n_prereleases_scipy_commits.__name__,
                    ]
                ],
                *[
                    (
                        repo_fixture_name,
                        None,
                    )
                    for repo_fixture_name in [
                        repo_w_github_flow_w_default_release_channel_conventional_commits.__name__,
                        repo_w_github_flow_w_default_release_channel_emoji_commits.__name__,
                        repo_w_github_flow_w_default_release_channel_scipy_commits.__name__,
                    ]
                ],
                *[
                    (
                        repo_fixture_name,
                        None,
                    )
                    for repo_fixture_name in [
                        repo_w_github_flow_w_feature_release_channel_conventional_commits.__name__,
                        repo_w_github_flow_w_feature_release_channel_emoji_commits.__name__,
                        repo_w_github_flow_w_feature_release_channel_scipy_commits.__name__,
                    ]
                ],
                *[
                    (
                        repo_fixture_name,
                        None,
                    )
                    for repo_fixture_name in [
                        repo_w_git_flow_w_alpha_prereleases_n_conventional_commits.__name__,
                        repo_w_git_flow_w_alpha_prereleases_n_emoji_commits.__name__,
                        repo_w_git_flow_w_alpha_prereleases_n_scipy_commits.__name__,
                    ]
                ],
                *[
                    (
                        repo_fixture_name,
                        None,
                    )
                    for repo_fixture_name in [
                        repo_w_git_flow_w_rc_n_alpha_prereleases_n_conventional_commits.__name__,
                        repo_w_git_flow_w_rc_n_alpha_prereleases_n_emoji_commits.__name__,
                        repo_w_git_flow_w_rc_n_alpha_prereleases_n_scipy_commits.__name__,
                    ]
                ],
                *[
                    (
                        repo_fixture_name,
                        "submod-v{version}",
                    )
                    for repo_fixture_name in [
                        repo_w_git_flow_w_rc_n_alpha_prereleases_n_conventional_commits_using_tag_format.__name__,
                    ]
                ],
            ]
        ],
    ],
)
def test_version_initializes_changelog_in_update_mode_w_no_prev_changelog(
    repo_result: BuiltRepoResult,
    repo_fixture_name: str,
    get_versions_from_repo_build_def: GetVersionsFromRepoBuildDefFn,
    tag_format: str,
    run_cli: RunCliFn,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    changelog_file: Path,
    stable_now_date: GetStableDateNowFn,
    get_cached_repo_data: GetCachedRepoDataFn,
):
    """
    Given that the changelog file does not exist,
    When the version command is run with changelog.mode set to "update",
    Then the version is created and the changelog file is initialized
    with the default content.
    """
    repo = repo_result["repo"]
    latest_tag = tag_format.format(
        version=get_versions_from_repo_build_def(repo_result["definition"])[-1]
    )

    if not (repo_build_data := get_cached_repo_data(repo_fixture_name)):
        pytest.fail("Repo build date not found in cache")

    repo_build_datetime = datetime.strptime(repo_build_data["build_date"], "%Y-%m-%d")
    now_datetime = stable_now_date().replace(
        year=repo_build_datetime.year,
        month=repo_build_datetime.month,
        day=repo_build_datetime.day,
    )

    # Capture the expected changelog content
    expected_changelog_content = changelog_file.read_text()

    # Reverse last release
    repo.git.tag("-d", latest_tag)
    repo.git.reset("--hard", "HEAD~1")

    # Set the project configurations
    update_pyproject_toml(
        "tool.semantic_release.changelog.mode", ChangelogMode.UPDATE.value
    )
    update_pyproject_toml(
        "tool.semantic_release.changelog.default_templates.changelog_file",
        str(changelog_file.name),
    )

    # Remove any previous changelog to update
    os.remove(str(changelog_file.resolve()))

    # Act
    with freeze_time(now_datetime.astimezone(timezone.utc)):
        cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, "--no-push", "--changelog"]
        result = run_cli(cli_cmd[1:])

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)

    # Check that the changelog file was re-created
    assert changelog_file.exists()

    actual_content = changelog_file.read_text()

    # Check that the changelog content is the same as before
    assert expected_changelog_content == actual_content


@pytest.mark.parametrize(
    "changelog_file, insertion_flag",
    [
        (
            lazy_fixture(example_changelog_md.__name__),
            lazy_fixture(default_md_changelog_insertion_flag.__name__),
        ),
        (
            lazy_fixture(example_changelog_rst.__name__),
            lazy_fixture(default_rst_changelog_insertion_flag.__name__),
        ),
    ],
)
@pytest.mark.usefixtures(repo_w_trunk_only_conventional_commits.__name__)
def test_version_maintains_changelog_in_update_mode_w_no_flag(
    changelog_file: Path,
    run_cli: RunCliFn,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    insertion_flag: str,
):
    """
    Given that the changelog file exists but does not contain the insertion flag,
    When the version command is run with changelog.mode set to "update",
    Then the version is created but the changelog file is not updated.
    """
    update_pyproject_toml(
        "tool.semantic_release.changelog.mode", ChangelogMode.UPDATE.value
    )
    update_pyproject_toml(
        "tool.semantic_release.changelog.default_templates.changelog_file",
        str(changelog_file.name),
    )

    # Remove the insertion flag from the existing changelog
    with changelog_file.open(newline=os.linesep) as rfd:
        expected_changelog_content = rfd.read().replace(
            f"{insertion_flag}{os.linesep}",
            "",
            1,
        )
    # no newline translations
    with changelog_file.open("w", newline="") as wfd:
        wfd.write(expected_changelog_content)

    # Act
    cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, "--no-push", "--changelog"]
    result = run_cli(cli_cmd[1:])

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)

    # Ensure changelog exists
    assert changelog_file.exists()

    # Capture the new changelog content (os aware because of expected content)
    with changelog_file.open(newline=os.linesep) as rfd:
        actual_content = rfd.read()

    # Check that the changelog content is the same as before
    assert expected_changelog_content == actual_content


@pytest.mark.parametrize(
    "changelog_file",
    [
        lazy_fixture(example_changelog_md.__name__),
        lazy_fixture(example_changelog_rst.__name__),
    ],
)
@pytest.mark.parametrize(
    "repo_result, repo_fixture_name, commit_type, tag_format",
    [
        (
            lazy_fixture(repo_fixture),
            repo_fixture,
            repo_fixture.split("_")[-2],
            "v{version}",
        )
        for repo_fixture in [
            # Must have a previous release/tag
            repo_w_trunk_only_conventional_commits.__name__,
        ]
    ],
)
def test_version_updates_changelog_w_new_version_n_filtered_commit(
    repo_result: BuiltRepoResult,
    repo_fixture_name: str,
    get_versions_from_repo_build_def: GetVersionsFromRepoBuildDefFn,
    commit_type: CommitConvention,
    tag_format: str,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    run_cli: RunCliFn,
    changelog_file: Path,
    stable_now_date: GetStableDateNowFn,
    get_commits_from_repo_build_def: GetCommitsFromRepoBuildDefFn,
    get_cached_repo_data: GetCachedRepoDataFn,
):
    """
    Given a project that has a version bumping change but also an exclusion pattern for the same change type,
    When the version command is run,
    Then the version is created and the changelog file is updated with the excluded commit
        info anyway.
    """
    repo = repo_result["repo"]
    latest_version = get_versions_from_repo_build_def(repo_result["definition"])[-1]
    latest_tag = tag_format.format(version=latest_version)
    repo_definition = get_commits_from_repo_build_def(repo_result["definition"])

    if not (repo_build_data := get_cached_repo_data(repo_fixture_name)):
        pytest.fail("Repo build date not found in cache")

    repo_build_datetime = datetime.strptime(repo_build_data["build_date"], "%Y-%m-%d")
    now_datetime = stable_now_date().replace(
        year=repo_build_datetime.year,
        month=repo_build_datetime.month,
        day=repo_build_datetime.day,
    )

    # expected version bump commit (that should be in changelog)
    bumping_commit = repo_definition[str(latest_version)]["commits"][-1]
    expected_bump_message = bumping_commit["desc"].capitalize()

    # Capture the expected changelog content
    expected_changelog_content = changelog_file.read_text()

    # Reverse last release
    repo.git.tag("-d", latest_tag)
    repo.git.reset("--hard", "HEAD~1")

    # Set the project configurations
    update_pyproject_toml(
        "tool.semantic_release.changelog.mode", ChangelogMode.UPDATE.value
    )
    update_pyproject_toml(
        "tool.semantic_release.changelog.default_templates.changelog_file",
        str(changelog_file.name),
    )
    update_pyproject_toml(
        "tool.semantic_release.changelog.exclude_commit_patterns",
        [f"{bumping_commit['msg'].split(':', maxsplit=1)[0]}: .*"],
    )

    # Act
    with freeze_time(now_datetime.astimezone(timezone.utc)):
        cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, "--no-push", "--changelog"]
        result = run_cli(cli_cmd[1:])

    # Capture the new changelog content (os aware because of expected content)
    actual_content = changelog_file.read_text()

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)
    assert expected_changelog_content == actual_content

    for msg_part in expected_bump_message.split("\n\n"):
        assert msg_part.capitalize() in actual_content



from __future__ import annotations

import json
import subprocess
from pathlib import Path
from shutil import rmtree
from textwrap import dedent
from typing import TYPE_CHECKING

import git
import pytest
from click.testing import CliRunner
from pytest_lazy_fixtures.lazy_fixture import lf as lazy_fixture

from semantic_release import __version__

from tests.const import MAIN_PROG_NAME, SUCCESS_EXIT_CODE, VERSION_SUBCMD
from tests.fixtures.repos import repo_w_no_tags_conventional_commits
from tests.util import assert_exit_code, assert_successful_exit_code

if TYPE_CHECKING:
    from tests.conftest import RunCliFn
    from tests.e2e.conftest import StripLoggingMessagesFn
    from tests.fixtures.example_project import ExProjectDir, UpdatePyprojectTomlFn
    from tests.fixtures.git_repo import BuiltRepoResult


@pytest.mark.parametrize(
    "project_script_name",
    [
        "python-semantic-release",
        "semantic-release",
        "psr",
    ],
)
def test_entrypoint_scripts(project_script_name: str):
    # Setup
    command = str.join(" ", [project_script_name, "--version"])
    expected_output = f"semantic-release, version {__version__}\n"

    # Act
    proc = subprocess.run(  # noqa: S602, PLW1510
        command, shell=True, text=True, capture_output=True
    )

    # Evaluate
    assert SUCCESS_EXIT_CODE == proc.returncode  # noqa: SIM300
    assert expected_output == proc.stdout
    assert not proc.stderr


def test_main_prints_version_and_exits(run_cli: RunCliFn):
    cli_cmd = [MAIN_PROG_NAME, "--version"]

    # Act
    result = run_cli(cli_cmd[1:])

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)
    assert result.output == f"semantic-release, version {__version__}\n"


def test_main_no_args_passes_w_help_text():
    from semantic_release.cli.commands.main import main

    cli_cmd = [MAIN_PROG_NAME]
    result = CliRunner().invoke(main, prog_name=cli_cmd[0])
    assert_successful_exit_code(result, cli_cmd)
    assert "Usage: " in result.output


@pytest.mark.parametrize(
    "repo_result",
    [lazy_fixture(repo_w_no_tags_conventional_commits.__name__)],
)
def test_not_a_release_branch_exit_code(
    repo_result: BuiltRepoResult, run_cli: RunCliFn
):
    # Run anything that doesn't trigger the help text
    repo_result["repo"].git.checkout("-b", "branch-does-not-exist")

    # Act
    cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, "--no-commit"]
    result = run_cli(cli_cmd[1:])

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)


@pytest.mark.parametrize(
    "repo_result",
    [lazy_fixture(repo_w_no_tags_conventional_commits.__name__)],
)
def test_not_a_release_branch_exit_code_with_strict(
    repo_result: BuiltRepoResult,
    run_cli: RunCliFn,
):
    # Run anything that doesn't trigger the help text
    repo_result["repo"].git.checkout("-b", "branch-does-not-exist")

    # Act
    cli_cmd = [MAIN_PROG_NAME, "--strict", VERSION_SUBCMD, "--no-commit"]
    result = run_cli(cli_cmd[1:])

    # Evaluate
    assert_exit_code(2, result, cli_cmd)


@pytest.mark.parametrize(
    "repo_result",
    [lazy_fixture(repo_w_no_tags_conventional_commits.__name__)],
)
def test_not_a_release_branch_detached_head_exit_code(
    repo_result: BuiltRepoResult,
    run_cli: RunCliFn,
):
    expected_err_msg = (
        "Detached HEAD state cannot match any release groups; no release will be made"
    )

    # cause repo to be in detached head state without file changes
    repo_result["repo"].git.checkout("HEAD", "--detach")

    # Act
    cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, "--no-commit"]
    result = run_cli(cli_cmd[1:])

    # detached head states should throw an error as release branches cannot be determined
    assert_exit_code(1, result, cli_cmd)
    assert expected_err_msg in result.stderr


@pytest.fixture
def toml_file_with_no_configuration_for_psr(tmp_path: Path) -> Path:
    path = tmp_path / "config.toml"
    path.write_text(
        dedent(
            r"""
            [project]
            name = "foo"
            version = "1.2.0"
            """
        )
    )

    return path


@pytest.fixture
def json_file_with_no_configuration_for_psr(tmp_path: Path) -> Path:
    path = tmp_path / "config.json"
    path.write_text(json.dumps({"foo": [1, 2, 3]}))

    return path


@pytest.mark.usefixtures(repo_w_no_tags_conventional_commits.__name__)
def test_default_config_is_used_when_none_in_toml_config_file(
    run_cli: RunCliFn,
    toml_file_with_no_configuration_for_psr: Path,
):
    cli_cmd = [
        MAIN_PROG_NAME,
        "--noop",
        "--config",
        str(toml_file_with_no_configuration_for_psr),
        VERSION_SUBCMD,
    ]

    # Act
    result = run_cli(cli_cmd[1:])

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)


@pytest.mark.usefixtures(repo_w_no_tags_conventional_commits.__name__)
def test_default_config_is_used_when_none_in_json_config_file(
    run_cli: RunCliFn,
    json_file_with_no_configuration_for_psr: Path,
):
    cli_cmd = [
        MAIN_PROG_NAME,
        "--noop",
        "--config",
        str(json_file_with_no_configuration_for_psr),
        VERSION_SUBCMD,
    ]

    # Act
    result = run_cli(cli_cmd[1:])

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)


@pytest.mark.usefixtures(repo_w_no_tags_conventional_commits.__name__)
def test_errors_when_config_file_does_not_exist_and_passed_explicitly(
    run_cli: RunCliFn,
):
    cli_cmd = [
        MAIN_PROG_NAME,
        "--noop",
        "--config",
        "somenonexistantfile.123.txt",
        VERSION_SUBCMD,
    ]

    # Act
    result = run_cli(cli_cmd[1:])

    # Evaluate
    assert_exit_code(2, result, cli_cmd)
    assert "does not exist" in result.stderr


@pytest.mark.usefixtures(repo_w_no_tags_conventional_commits.__name__)
def test_errors_when_config_file_invalid_configuration(
    run_cli: RunCliFn,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    strip_logging_messages: StripLoggingMessagesFn,
    pyproject_toml_file: Path,
):
    # Setup
    update_pyproject_toml("tool.semantic_release.remote.type", "invalidType")
    cli_cmd = [MAIN_PROG_NAME, "--config", str(pyproject_toml_file), VERSION_SUBCMD]

    # Act
    result = run_cli(cli_cmd[1:])

    # preprocess results
    stderr_lines = strip_logging_messages(result.stderr).splitlines()

    # Evaluate
    assert_exit_code(1, result, cli_cmd)
    assert "1 validation error for RawConfig" in stderr_lines[0]
    assert "remote.type" in stderr_lines[1]


def test_uses_default_config_when_no_config_file_found(
    run_cli: RunCliFn,
    example_project_dir: ExProjectDir,
    change_to_ex_proj_dir: None,
):
    # We have to initialise an empty git repository, as the example projects
    # all have pyproject.toml configs which would be used by default
    with git.Repo.init(example_project_dir) as repo:
        rmtree(str(Path(repo.git_dir, "hooks")))

        repo.git.branch("-M", "main")

        with repo.config_writer("repository") as config:
            config.set_value("user", "name", "semantic release testing")
            config.set_value("user", "email", "not_a_real@email.com")
            config.set_value("commit", "gpgsign", False)
            config.set_value("tag", "gpgsign", False)

        repo.create_remote(name="origin", url="foo@barvcs.com:user/repo.git")
        repo.git.commit("-m", "feat: initial commit", "--allow-empty")

        cli_cmd = [MAIN_PROG_NAME, "--noop", VERSION_SUBCMD]

        # Act
        result = run_cli(cli_cmd[1:])

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)
